/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO

#include <stdio.h> /* siscanf */

#include <wlp_api.h>
#include <wlp_inet.h>

#include <owl/core/owl_debug.h>
#include <owl/drivers/owl_led.h>

#include <pico/wlp_httpd.h>

#include "web_logo.h"
#include "tempsensor.h"


/*!
 * \file web_content.c
 * \brief wlp_httpd web server content
 *
 * The following steps are needed to add pages to the web server
 * 
 * 1. Add the name of the new page to \a http_get()
 * 2. Implement a file access function in web_content.c. The function
 *    should be called from \a http_get() and should use the local
 *    reply_str() or reply_buf() functions to write the contents of the file
 *    (HTML page) to the web browser at the other end of the connection.
 * 3. If the new page contains HTTP form elements, add a function
 *    to web_contents.c to handle the form fields and call it from
 *    \a http_get() if the request path matches the file name
 *    with an appended '?' (see the handling of "index.html?" as
 *    an example).
 *
 */

static int reply_str(int sock, const char *buf);
static int reply_buf(int sock, const char *buf, int len);

/* Use progmem on AVR8 to save ram memory since we can't keep const strings
 * in flash without using the PROGMEM or PSTR attribute
 */
#ifdef __avr8__
# include <avr/pgmspace.h>
static int reply_str_P(int sock, const char *buf);
static int reply_buf_P(int sock, const char *buf, int len);
#define HTTP_BUFSIZE 64

/* On linux, we dont use any PROGMEM stuff */
#else
# define PSTR(str) str
# define reply_str_P reply_str
# define reply_buf_P reply_buf
#endif /* __avr8__ */



static int file_index(int sock);
static int file_logo(int sock);
static int file_404(int sock);
static int file_index_form_field(int sock, char *fields);


/* Lookup the file specified by path and send it as a
 * reply on the connected socket. */
int http_get(char *path, int sock)
{

        owl_dbg("path:%s", path);
        
        /* The path "/" is and alias for "index.html" */
        if (owl_os_strcmp(path, "/") == 0 || owl_os_strcmp(path, "/?") == 0) {
                return file_index(sock);

        /* If we have a trailing ? character we expect
         * the rest of the path to be a list of form fields
         * on the form "<name>=<value>?<name>=<value>..."
         * These fields are handled by a special function.
         */
        } else if (owl_os_strncmp(path, "/index.html?", 12) == 0 ) {
                return file_index_form_field(sock, path + 12);
        } else if (owl_os_strcmp(path, "/index.html") == 0) {
                return file_index(sock);
        } else if (owl_os_strcmp(path, "/logo.gif") == 0) {
                return file_logo(sock);
        }
        /* If the path didn't match any supported file then
         * we return a HTTP 404 error page. */
        else {
                return file_404(sock);
        }
        return -1;
}


/* Handle form fields on the index page. */
static int file_index_form_field(int sock, char *fields) {
        int led_index;
        owl_info("Field \"%s\"", fields);
        if ( 0 == siscanf(fields, "toggle=LED%d", &led_index) ) {
                return file_index(sock);
        }
        switch (led_index) {
        case 0: owlled_toggle(OWL_BOARD_LED0); break;
        case 1: owlled_toggle(OWL_BOARD_LED1); break;
        case 2: owlled_toggle(OWL_BOARD_LED2); break;
        case 3: owlled_toggle(OWL_BOARD_LED3); break;
        case 4: owlled_toggle(OWL_BOARD_LED4); break;
        case 5: owlled_toggle(OWL_BOARD_LED5); break;
        case 6: owlled_toggle(OWL_BOARD_LED6); break;
        case 7: owlled_toggle(OWL_BOARD_LED7); break;
        default: break;
        };
        
        return file_index(sock);
}


static int file_logo(int sock)
{
        int err;
        
        /* http header */
        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 200 OK\n"
                          "Server: wlp_httpd\n"
                          "Content-type: image/gif\n\n"))) < 0)
                return err;

        /* image data */
        if ((err = reply_buf_P(sock, web_logo, sizeof(web_logo))) < 0)
                return err;

        return 0;
}

static int file_index(int sock)
{
        int err;
        struct ip_addr ip;
        struct wlp_mac_addr mac;
        char macstr[18] = "N/A";
        char *ipstr = "N/A";
#if 0
        int temp;
        char tempstr[16];
#endif
        
        if ( WLP_INVALID_SOCKET_ID == sock ) {
                return -1;
        }

        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 200 OK\n"
                          "Server: wlp_httpd\n"
                          "Content-type: text/html\n\n"
                          "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 "
                          "Transitional//EN\" "
                          "\"http://www.w3.org/TR/html4/loose.dtd\">"
                          "<html>"
                          "<head><title>oWL pico HTTP Web server</title>"
                          "</head>"
                          "<body style=font-family:arial>"
                          "<table><tr><td width=\"100\"></td><td>"
                          "<table><tr valign=\"top\"><td>"
                          "<img src=\"logo.gif\" alt=\"logo\" "
                          "title=\"logo\">"
                          "</td></tr>"
                          "<tr><td width=\"500\">"
                          "<h2>Web Server running on oWL pico</h2>"
                          "<p>"
                          "The web page is served by wlp_httpd using "
                          "the H&D Wireless WLAN module and the "
                          "oWL-pico&copy; Software API; "
                          "a unique Wi-Fi solution from <b>H&amp;D Wireless AB</b>."
                          "</p><hr><table>"))) < 0)
                return err;

        if (wlp_get_ipaddr(&ip, NULL, NULL, NULL) == 0)
                ipstr = inet_ntoa(&ip);
        
        if ((err = reply_str_P(sock, PSTR("<tr><td>IP Address: "))) < 0)
                return err;
        
        if ((err = reply_str(sock, ipstr)) < 0)
                return err;
        
        if ((err = reply_str_P(sock, PSTR("</td><td><b></b></td></tr>"))) < 0)
                return err;
        
        if (wlp_get_hwaddr(&mac) == 0)
                sniprintf(macstr, sizeof(macstr),
                          "%02x:%02x:%02x:%02x:%02x:%02x",
                          mac.octet[0], mac.octet[1], mac.octet[2],
                          mac.octet[3], mac.octet[4], mac.octet[5]);
        
        if ((err = reply_str_P(sock, PSTR("<tr><td>MAC Address: "))) < 0)
                return err;

        if ((err = reply_str(sock, macstr)) < 0)
                return err;

        if ((err = reply_str_P(sock, PSTR("</td><td><b></b></td></tr>"))) < 0)
                return err;

#if 0 /* temperature */
        if ((err = reply_str_P(sock, PSTR("<tr><td>Temperature: "))) < 0)
                return err;

#if defined(__avr8__)
        temp = tempsensor_get_temp();
#else
        temp = 23;
#endif

        sniprintf(tempstr, sizeof tempstr, "%d C", temp);
        if ((err = reply_str(sock, tempstr)) < 0)
            return err;
        
        if ((err = reply_str_P(sock, PSTR("</td><td><b></b></td></tr>"))) < 0)
                return err;
#endif /* temperature */
        if ((err = reply_str_P(
                     sock,
                     PSTR("<tr><td><form method=\"get\" action=\"\">"
                          "<p><input type=\"submit\" style=\"height: "
                          "50px\" value=\"Refresh\">"
                          "</p></form></td></tr></table><br>"
                          "<form id=\"form\""
                          "name=\"form\" method=\"get\" "
                          "action=\"index.html\">"
                          "<fieldset>"
                          "<legend>LED control</legend>"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED0\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED1\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED2\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED3\">&nbsp&nbsp<p>"))) < 0) {
                return err;
        }

        if ((err = reply_str_P(
                     sock,
                     PSTR("<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED4\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED5\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED6\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED7\">&nbsp&nbsp"
                          "</fieldset></form></td>"
                          "<td>&nbsp;</td></tr></table>"
                          "</td><td width=\"20\%\"></td></tr></table>"
                          "</body></html>"))) < 0) {
                return err;
        }

        return 0;
}

/* Write a HTTP 404 error page to the socket. */
static int file_404(int sock) {
        int err;
        
        if ( WLP_INVALID_SOCKET_ID == sock ) {
                return -1;
        }

        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 404 File not found\n"
                          "Server: wlp_httpd\n"
                          "Content-type: text/html\n\n"
                          "<html>"
                          "<head><title>oWL pico HTTP Web server</title></head>"
                          "<body>"
                          "<h2>404 - Page not found</h2>"
                          "<p>"
                          "Sorry, the page you are requesting was not found on "
                          "this server."
                          "</p></body></html>"))) < 0)
                return err;

        return 0;
}


/* Write the specified buffer to the socket, looping until
 * the whole buffer has been sent. */
static int reply_buf(int sock, const char *buf, int len)
{
        int err;
        int pos = 0;
        while (pos < len) {
                if ((err = wlp_send(sock, buf + pos, len - pos)) < 0)
                        return err;

                pos += err;
        }
        
        return 0;
}


/* Convenience function for writint a NULL terminated string to the socket */
static int reply_str(int sock, const char *str)
{
    return reply_buf(sock, str, owl_os_strlen(str));
}



#ifdef __avr8__

/* Write the specified PROGMEM buffer to the socket using the wlp_send()
 * method in the wlp_api, looping until the whole buffer has been sent.
 *
 * This function is specific for avr8 architectures when sending data that
 * lives in the internal flash memory.
 */
static int reply_buf_P(int sock, const char *buf, int len)
{
        char *rambuf;
        int pos = 0;

        if ((rambuf = owl_os_alloc(HTTP_BUFSIZE)) < 0)
                return -1;
        
        while (pos < len) {
                int err;
                int chunklen = len - pos;

                /* copy chunk of data from progmem to ram buffer */
                if (chunklen > HTTP_BUFSIZE)
                        chunklen = HTTP_BUFSIZE;
                memcpy_P(rambuf, buf + pos, chunklen);

                /* send the chunk */
                if ((err = wlp_send(sock, rambuf, chunklen)) < 0) {
                        owl_os_free(rambuf);
                        return err;
                }

                /* update input buffer position by the number of bytes
                 * actually sent.
                 */
                pos += err;
        }
        owl_os_free(rambuf);
        return 0;
}

/* Convenience function for writint a NULL terminated string to the socket */
static int reply_str_P(int sock, const char *str)
{
        return reply_buf_P(sock, str, strlen_P(str));
} 

#endif /* __avr8__ */



