/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG

#include <wlp_api.h>
#include <wlp_inet.h>

#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_defs.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_rtc.h>
#include <owl/drivers/owl_led.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/components/spb800p.h>
#include <pico/wlp_shell.h>
#include <pico/wlp_httpd.h>

#include "config.h"

#if defined(__linux__)
# include <linux/util/mtdfile.h>
# define OWL_BOARD_SPB800P_UART ((void *) fd)
#endif

#define TIMER_PERIOD_MS 64

struct httpd *http_handle = NULL;
void *owluart = NULL;

static void link_cb(void *ctx, int link)
{
        owl_info("link:%d", link);
}

static void addr_cb(void *ctx, const struct ip_addr *ip)
{
        static struct httpd *httpd = NULL;
        struct owlsh_ctx *sh = ctx;
        
        if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
                if (httpd) {
                        owlsh_printf(sh, "stopping httpd\n\r");
                        httpd_destroy(httpd);
                        httpd = NULL;
                }

        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", inet_ntoa(ip));
                owlsh_printf(sh, "starting httpd\n\r");
                if ((httpd = httpd_create(80)) == NULL)
                        owl_err("fail to start httpd");
        }
}


static void tick_cb(void)
{
        uint32_t period;
        period = owlrtc_get_period();
        owltmr_tick(period);
}

#if defined(OWL_BOARD_SPB800P_WAKEUP_PIN)
static void wakeup_cb(void *ctx, int enable) {

        owl_dbg("wakeup enabled == %d", enable);
        if (enable) {
                owlgpio_clr(OWL_BOARD_SPB800P_WAKEUP_PIN);
#if defined (OWL_BOARD_SPB800P_HOST_ATTENTION_PIN) && defined (HA_ACK)
                /* Wait until the SPB800 signals that it's ready. */
                while (owlgpio_is_high(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN));
#else
                /* no host attention pin, hope that 5ms is enough */
                owlboard_mdelay(30);
#endif
        }
        else
                owlgpio_set(OWL_BOARD_SPB800P_WAKEUP_PIN);
}
#endif /* OWL_BOARD_SPB800P_WAKEUP_PIN */

static void select_cb(void *uart, int select)
{
# ifdef __avr8__
        static irqflags_t iflags;
        (void) uart;
        
        if (select)
                iflags = cpu_irq_save();
        else
                cpu_irq_restore(iflags);
# else
        return;
# endif
}

#if defined(__linux__) /* XXX */
int
owlboard_init(void)
{
        return 0;
}

struct owlboard_data *
owlboard_data(void)
{
        struct owlboard_data *bdata;        
        static struct owlboard_data xdata;

        bdata = &xdata;
        if ((bdata->mtd = mtdfile_create("mtd")) == NULL) {
                owl_err("fail to init mtdfile");
                return NULL;
        }

        return bdata;
}
#endif

int main(int argc, char *argv[])
{
        int err;
        struct owlsh_ctx *sh;
        struct config cfg;
        int shflags = OWLSH_ESCAPECHARS;
        struct owlboard_data *bdata;
        
#if defined(__linux__)
        int fd;
        char *devname;

        shflags |= OWLSH_NOECHO;
        
        if (argc > 1 && !owl_os_strcmp(argv[1], "-h")) {
                printf("%s [port]\n", argv[0]);
                return 0;
        }

        if(argc > 1)
                devname = argv[1];
        else
                devname = "/dev/ttyUSB0";


        if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
                owl_err("Failed to open device '%s'", devname);
                return -1;
        }
#endif
        owlboard_init();
        bdata = owlboard_data();
        owlrtc_init(TIMER_PERIOD_MS, tick_cb);

        if ((err = owltmr_init()) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }
        
        /* We will use the owl_* portability layer for console
         * input/output and error/debug functionality. This allows us
         * to build this code on (for instance) a Linux PC system
         * during development. To build the code for the target 
         * host platform you need to 
         *  - Write a owl_env.h file for the host
         *  - Write a owl_uart.c file for the target platform, implementing
         *    the functions prototyped in owl_uart.h
         *  - Write a owl_led.c file for the target platform, implementing
         *    the functions prototyped in owl_led.h
         *  - Write a owl_rtc.c file for the target platform, implementing
         *    the functions prototyped in owl_rtc.h
         */

        /* Initialize the console UART */
        if (owluart_init(OWL_BOARD_SHELL_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;
        
        /* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(NULL, owluart_write, OWL_BOARD_SHELL_UART,
                               shflags)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }
        
        /* Power on SPB800 */
        spb800p_resume();

        /* Initialize the host-WiFi UART */
        if ((err = owluart_init(OWL_BOARD_SPB800P_UART, UART_BAUDRATE_115200,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init pico err:%d(%s)", err, owl_error(err));
                return -1;
        }

        
        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(WLP_INTERFACE_UART,
                            owluart_read, owluart_write, select_cb,
                            OWL_BOARD_SPB800P_UART)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* keep track of the uart fd to support the set_baudrate command */
        owluart = OWL_BOARD_SPB800P_UART;

#if defined(OWL_BOARD_SPB800P_WAKEUP_PIN)
        /* Register a wakeup pin control callback.
         * This is used to wake the SPB800 when Power save level 1 is
         * used. */
        wlp_set_wakeup_cb(wakeup_cb, sh);
#endif /* defined(OWL_BOARD_SPB800P_WAKEUP_PIN) */

        
        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_link_cb(link_cb, NULL);
        wlp_set_ipaddr_cb(addr_cb, sh);

        /* Add misc commands */
        owlsh_addcmd(sh, &cmd_fw_upgrade, NULL);
        owlsh_addcmd(sh, &cmd_get_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_version, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_config, bdata->mtd);
        owlsh_addcmd(sh, &cmd_help, NULL);


        if ((err = config_read(bdata->mtd, &cfg))) {
                owl_info("could not read config, fall back to default");
                config_default(&cfg);
        }
        if ((err = config_apply(&cfg)))
                owl_err("could not apply configuration");
        
        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        for (;;) {
                if (owluart_poll(OWL_BOARD_SHELL_UART))
                        owlsh_key(sh, owluart_inb(OWL_BOARD_SHELL_UART, 0));

#if defined(__avr8__)
                if (ioport_pin_is_low(GPIO_PUSH_BUTTON_7)) {
                        if (config_is_enabled()) {

                                if ((err = config_read(bdata->mtd, &cfg))) {
                                        owl_info("could not read config, "
                                                 "fall back to default");
                                        config_default(&cfg);
                                }
                                
                                if ((err = config_apply(&cfg)))
                                        owl_err("could not apply "
                                                "configuration");

                                
                        } else {
                                if ((err = config_enable()))
                                        owl_err("could not enable config");
                        }
                }
#endif

                wlp_poll();
                owltmr_poll();
        }                
        
        return 0;
}

/*!  @} */
