/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <wlp_api.h>
#include <wlp_inet.h>

#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>
#include <owl/util/querystring.h>
#include <owl/util/mtd/mtd.h>
#include <owl/drivers/owl_led.h>

#include <pico/wlp_httpd.h>
#include "web_logo.h"
#include "config.h"



/*!
 * \file web_content.c
 * \brief wlp_httpd web server content
 *
 * The following steps are needed to add pages to the web server
 * 
 * 1. Add the name of the new page to \a http_get()
 * 2. Implement a file access function in web_content.c. The function
 *    should be called from \a http_get() and should use the local
 *    reply_str() or reply_buf() functions to write the contents of the file
 *    (HTML page) to the web browser at the other end of the connection.
 * 3. If the new page contains HTTP form elements, add a function
 *    to web_contents.c to handle the form fields and call it from
 *    \a http_get() if the request path matches the file name
 *    with an appended '?' (see the handling of "index.html?" as
 *    an example).
 *
 */

static int reply_str(int sock, const char *buf);
static int reply_buf(int sock, const char *buf, int len);

#define HTTP_BUFSIZE 256

/* Use progmem on AVR8 to save ram memory since we can't keep const strings
 * in flash without using the PROGMEM or PSTR attribute
 */
#ifdef __avr8__
# include <avr/pgmspace.h>
static int reply_str_P(int sock, const char *buf);
static int reply_buf_P(int sock, const char *buf, int len);

/* On linux, we dont use any PROGMEM stuff */
#else
# define PSTR(str) str
# define reply_str_P reply_str
# define reply_buf_P reply_buf
#endif /* __avr8__ */

static int reply_fmt(int sock, const char *format, ...);

int locked = 0;

static int file_index(int sock);
static int file_control(int sock);
static int file_config(int sock);
static int file_logo(int sock);
static int file_404(int sock);
static void parse_form_field(int sock, char *fields);


/* Lookup the file specified by path and send it as a
 * reply on the connected socket. */
int http_get(char *path, int sock)
{

        owl_dbg("path:%s", path);
        
        /* The path "/" is and alias for "index.html" */
        if (owl_os_strcmp(path, "/") == 0 || owl_os_strcmp(path, "/?") == 0) {
                return file_index(sock);
        } 
        /* If we have a trailing ? character we expect
         * the rest of the path to be a list of form fields
         * on the form "<name>=<value>?<name>=<value>..."
         * These fields are handled by a special function.
         */
        else if (owl_os_strncmp(path, "/index.html?", 12) == 0) {
                parse_form_field(sock, path + 12);
                return file_index(sock);
        } else if (owl_os_strcmp(path, "/index.html") == 0 ) {
                return file_index(sock);
        } else if (owl_os_strcmp(path, "/logo.gif") == 0 ) {
                return file_logo(sock);
        }
        /* If the path didn't match any supported file then
         * we return a HTTP 404 error page. */
        else {
                return file_404(sock);
        }
        return -1;
}


static void
apply_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct config cfg;
        int err;
        struct mtd_dev *mtd = ctx;
        
        locked = 0;

        if ((err = config_read(mtd, &cfg))) {
                owl_err("fail to read config err:%d", err);
                return;
        }
        
        if ((err = config_apply(&cfg))) {
                owl_err("fail to apply config err:%d", err);
                return;
        }

        owltmr_deltmo(tmo);

        owl_info("configuration applied");
}


/* Handle form fields on the index page. */
static void
parse_form_field(int sock, char *fields)
{
        struct querystring q[9];
        struct config cfg;
        int cnt;
        int led_index;
        int err;
        char *p;
        struct mtd_dev *mtd = owlboard_data()->mtd;

        owl_info("field:\'%s\'", fields);

        
        if ((cnt = querystring_parse(fields, q, owl_array_size(q))) <= 0) {
                owl_err("no valid querystring");
                return;
        }

        if (!config_is_enabled()) {
        
                if ((p = querystring_lookup("toggle", q, cnt)) == NULL)
                        return;

                if (owl_os_strncmp(p, "LED", 3))
                        return;
                
                led_index = owl_os_strtol(p + 3, NULL, 10);
                switch (led_index) {
                case 0: owlled_toggle(OWL_BOARD_LED0); break;
                case 1: owlled_toggle(OWL_BOARD_LED1); break;
                case 2: owlled_toggle(OWL_BOARD_LED2); break;
                case 3: owlled_toggle(OWL_BOARD_LED3); break;
                case 4: owlled_toggle(OWL_BOARD_LED4); break;
                case 5: owlled_toggle(OWL_BOARD_LED5); break;
                case 6: owlled_toggle(OWL_BOARD_LED6); break;
                case 7: owlled_toggle(OWL_BOARD_LED7); break;
                default: break;
                };

                return;
        }

        config_default(&cfg);
        
        if ((p = querystring_lookup("p0", q, cnt)) == NULL)
                return;
        strncpy(cfg.ssid.ssid, p, sizeof(cfg.ssid.ssid));
        cfg.ssid.len = strlen(p);
        
        if ((p = querystring_lookup("p1", q, cnt)) == NULL)
                return;
        strncpy(cfg.key, p, sizeof(cfg.key));
        cfg.key[sizeof(cfg.key) - 1] = 0;
        
        if ((p = querystring_lookup("p2", q, cnt)) == NULL)
                return;
        cfg.wep = atoi(p);
        
        if ((p = querystring_lookup("p3", q, cnt)) == NULL)
                return;
        cfg.dhcp = atoi(p);
        
        if ((p = querystring_lookup("p4", q, cnt)) == NULL)
                return;
        if (!inet_aton(p, &cfg.ip))
                return;
        
        if ((p = querystring_lookup("p5", q, cnt)) == NULL)
                return;
        if (!inet_aton(p, &cfg.mask))
                return;
        
        if ((p = querystring_lookup("p6", q, cnt)) == NULL)
                return;
        if (!inet_aton(p, &cfg.gw))
                return;
        
        if ((p = querystring_lookup("p7", q, cnt)) == NULL)
                return;
        if (!inet_aton(p, &cfg.dns))
                return;
        
        owl_info("writing configuration");
        if ((err = config_write(mtd, &cfg)) < 0) {
                owl_err("fail to write config err:%d", err);
                return;
        }
        
        /* Commit the config after giving the http server a
           while to return this page to the peer. Then reset
           the device */
        owl_info("scheduling apply");
        if (owltmr_addtmo(5000, 0, apply_cb, mtd) == NULL) {
                owl_err("failed to add apply_cb to tmr list");
                return;
        }
        
        locked = 1;
}
                

static int file_logo(int sock)
{
        int err;
        
        /* http header */
        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 200 OK\n"
                          "Server: wlp_httpd\n"
                          "Content-type: image/gif\n\n"))) < 0)
                return err;

        /* image data */
        if ((err = reply_buf_P(sock, web_logo, sizeof(web_logo))) < 0)
                return err;

        return 0;
}

static int file_control(int sock)
{
        int err;
        struct ip_addr ip;
        struct wlp_mac_addr mac;
        char macstr[18] = "N/A";
        char *ipstr = "N/A";
        
        if ( WLP_INVALID_SOCKET_ID == sock ) {
                return -1;
        }

        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 200 OK\n"
                          "Server: wlp_httpd\n"
                          "Content-type: text/html\n\n"
                          "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 "
                          "Transitional//EN\" "
                          "\"http://www.w3.org/TR/html4/loose.dtd\">"
                          "<html>"
                          "<head><title>oWL pico HTTP Web server</title>"
                          "</head>"
                          "<body style=font-family:arial>"
                          "<table><tr><td width=\"100\"></td><td>"
                          "<table><tr valign=\"top\"><td>"
                          "<img src=\"logo.gif\" alt=\"logo\""
                          "title=\"logo\">"
                          "</td></tr>"
                          "<tr><td width=\"500\">"
                          "<h2>Web Server running on oWL pico</h2>"
                          "<p>"
                          "The web page is served by wlp_httpd using "
                          "the SPB800 WLAN module and the "
                          "oWL-pico&copy; Software API; "
                          "a unique Wi-Fi solution from <b>H&amp;D Wireless AB</b>."
                          "</p><hr><table>"))) < 0)
                return err;

        if (wlp_get_ipaddr(&ip, NULL, NULL, NULL) == 0)
                ipstr = inet_ntoa(&ip);
        
        if ((err = reply_str_P(sock, PSTR("<tr><td>IP Address: "))) < 0)
                return err;
        
        if ((err = reply_str(sock, ipstr)) < 0)
                return err;
        
        if ((err = reply_str_P(sock, PSTR("</td><td><b></b></td></tr>"))) < 0)
                return err;
        
        if (wlp_get_hwaddr(&mac) == 0)
                sniprintf(macstr, sizeof(macstr),
                          "%02x:%02x:%02x:%02x:%02x:%02x",
                          mac.octet[0], mac.octet[1], mac.octet[2],
                          mac.octet[3], mac.octet[4], mac.octet[5]);
        
        if ((err = reply_str_P(sock, PSTR("<tr><td>MAC Address: "))) < 0)
                return err;

        if ((err = reply_str(sock, macstr)) < 0)
                return err;

        if ((err = reply_str_P(sock, PSTR("</td><td><b></b></td></tr>"))) < 0)
                return err;

        if ((err = reply_str_P(
                     sock,
                     PSTR("</table><br>"
                          "<form id=\"form\""
                          "name=\"form\" method=\"get\" "
                          "action=\"index.html\">"
                          "<fieldset>"
                          "<legend>LED control</legend>"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED0\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED1\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED2\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED3\">&nbsp&nbsp<p>"))) < 0) {
                return err;
        }

        if ((err = reply_str_P(
                     sock,
                     PSTR("<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED4\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED5\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED6\">&nbsp&nbsp"
                          "<input type=\"submit\" style=\"height: "
                          "50px\" name=\"toggle\" "
                          "value=\"LED7\">&nbsp&nbsp"
                          "</fieldset></form></td>"
                          "<td>&nbsp;</td></tr></table>"
                          "</td><td width=\"20\%\"></td></tr></table>"
                          "</body></html>"))) < 0) {
                return err;
        }

        return 0;
}

static int file_index(int sock)
{
        if (config_is_enabled())
                return file_config(sock);
        else
                return file_control(sock);
}
        


static int
html_add_message(int sock, const char *msg)
{
        return reply_fmt(sock, "<h2><font color=red>%s</font></h2><p>", msg);
}

static int
html_begin_form(int sock)
{
        return reply_str_P(sock, PSTR("<form method=get action=index.html>"));
}

static int
html_end_form(int sock)
{
        return reply_str_P(sock, PSTR("</form>"));
}

static int
html_begin_section(int sock, const char *desc)
{
        return reply_fmt(
                sock, "<fieldset><legend>%s</legend><table cellpadding=5>",
                desc);
}

static int
html_end_section(int sock)
{
        return reply_str_P(sock, PSTR("</table></fieldset>"));
}


static int
html_add_input(int sock, const char *desc, const char *name, const char *value)
{
        return reply_fmt(
                sock,
                "<tr><td>%s</td><td><input name=%s value=\"%s\" </td></tr>",
                desc, name, value);
}

static int
html_add_checkbox(int sock, const char *desc, const char *name, int checked)
{
        return reply_fmt(
                sock,
                "<tr><td>%s</td><td><input type=hidden name=%s value=0>"
                "<input type=checkbox name=%s value=1 %s></td><td></tr>",
                desc, name, name, checked ? "checked" : "");

}

static int
html_add_submit(int sock, const char *name, const char *value)
{
        return reply_fmt(
                sock,
                "<tr><td><input type=submit name=%s value=%s></td><td></tr>",
                name, value);
}


#if 0 /* unused */
static int
html_begin_radio(int sock, const char *desc)
{
        return reply_fmt(
                sock,
                "<tr><td>%s</td><td><table border=1 cellspacing=0 "
                "cellpadding=5><tr>", desc);
}


static int
html_add_radio_choice(int sock, const char *desc, const char *name,
                      const char *value, const char *checked)
{
        return reply_fmt(
                sock,
                "<td><label><input type=radio name=%s value=%s %s>%s"
                "</label></td>", name, value,
                !strcmp(checked, value) ? "checked" : "",
                desc);
}


static int
html_end_radio(int sock)
{
        return reply_str_P(sock, PSTR("</tr></table></td></tr>"));
}

static int
html_add_read_only(int sock, const char *desc, const char *value)
{
        return reply_fmt(
                sock,
                "<tr><td>%s</td><td><i>%s</i></td><td></tr>", desc, value);
}
#endif /* unused */

static int file_config(int sock)
{
        int err;
        struct config cfg;
        struct mtd_dev *mtd = owlboard_data()->mtd;
        
        if (config_read(mtd, &cfg) < 0)
                config_default(&cfg);
        
        if (sock == WLP_INVALID_SOCKET_ID)
                return -1;

        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 200 OK\n"
                          "Server: wlp_httpd\n"
                          "Content-type: text/html;charset=utf-8\n\n"
                          "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 "
                          "Transitional//EN\" "
                          "\"http://www.w3.org/TR/html4/loose.dtd\">"
                          "<html>"
                          "<head><title>oWL pico HTTP Web server</title>"
                          "</head>"
                          "<body style=font-family:arial>"
                          "<table><tr><td width=\"100\"></td><td>"
                          "<table><tr valign=\"top\"><td>"
                          "<img src=\"logo.gif\" alt=\"logo\""
                          "title=\"logo\">"
                          "</td></tr>"
                          "<tr><td width=\"500\">"
                          "<h1>Application Configuration</h1>"))) < 0)
                return err;

        if (locked) {
                if ((err = html_add_message(
                             sock,
                             "Writing configuration, "
                             "networking will reset!")) < 0)
                        return err;
        }
        
        if ((err = html_begin_form(sock)) < 0)
                return err;

        if ((err = html_begin_section(sock, "Wi-Fi Configuration")) < 0)
                return err;
        if ((err = html_add_input(sock, "SSID", "p0", ssid_str(&cfg.ssid))) < 0)
                return err;
        if ((err = html_add_input(sock, "Security Key", "p1", cfg.key)) < 0)
                return err;
        if ((err = html_add_checkbox(sock, "WEP", "p2", cfg.wep)) < 0)
                return err;
        if ((err = html_end_section(sock)) < 0)
                return err;

        if ((err = html_begin_section(sock, "IP Configuration")) < 0)
                return err;
        if ((err = html_add_checkbox(sock, "Enable DHCP", "p3", cfg.dhcp)) < 0)
                return err;
        if ((err = html_add_input(
                     sock, "IP Address", "p4", inet_ntoa(&cfg.ip))) < 0)
                return err;
        if ((err = html_add_input(
                     sock, "Network Mask", "p5", inet_ntoa(&cfg.mask))) < 0)
                return err;
        if ((err = html_add_input(
                     sock, "Gateway", "p6", inet_ntoa(&cfg.gw))) < 0)
                return err;
        if ((err = html_add_input(
                     sock, "DNS Server", "p7", inet_ntoa(&cfg.dns))) < 0)
                return err;
        if ((err = html_end_section(sock)) < 0)
                return err;

        if ((err = html_begin_section(sock, "Application Configuration")) < 0)
                return err;
        if ((err = reply_str_P(sock, PSTR("...<br>"))) < 0)
                return err;
        if ((err = html_end_section(sock)) < 0)
                return err;

        if ((err = html_add_submit(sock, "submit", "Submit")))
                return err;
        
        if ((err = html_end_form(sock)) < 0)
                return err;
        
        if ((err = reply_str_P(sock, PSTR("</body></html>"))) < 0)
                return err;

        return 0;
}


/* Write a HTTP 404 error page to the socket. */
static int file_404(int sock) {
        int err;
        
        if ( WLP_INVALID_SOCKET_ID == sock ) {
                return -1;
        }

        if ((err = reply_str_P(
                     sock,
                     PSTR("HTTP/1.0 404 File not found\n"
                          "Server: wlp_httpd\n"
                          "Content-type: text/html\n\n"
                          "<html>"
                          "<head><title>oWL pico HTTP Web server</title></head>"
                          "<body>"
                          "<h2>404 - Page not found</h2>"
                          "<p>"
                          "Sorry, the page you are requesting was not found on "
                          "this server."
                          "</p></body></html>"))) < 0)
                return err;

        return 0;
}


/* Write the specified buffer to the socket, looping until
 * the whole buffer has been sent. */
static int reply_buf(int sock, const char *buf, int len)
{
        int err;
        int pos = 0;
        while (pos < len) {
                if ((err = wlp_send(sock, buf + pos, len - pos)) < 0)
                        return err;

                pos += err;
        }
        
        return 0;
}


/* Convenience function for writint a NULL terminated string to the socket */
static int reply_str(int sock, const char *str)
{
        return reply_buf(sock, str, strlen(str));
}



#ifdef __avr8__

/* Write the specified PROGMEM buffer to the socket using the wlp_send()
 * method in the wlp_api, looping until the whole buffer has been sent.
 *
 * This function is specific for avr8 architectures when sending data that
 * lives in the internal flash memory.
 */
static int reply_buf_P(int sock, const char *buf, int len)
{
        char *rambuf;
        int pos = 0;

        if ((rambuf = malloc(HTTP_BUFSIZE)) < 0)
                return -1;
        
        while (pos < len) {
                int err;
                int chunklen = len - pos;

                /* copy chunk of data from progmem to ram buffer */
                if (chunklen > HTTP_BUFSIZE)
                        chunklen = HTTP_BUFSIZE;
                memcpy_P(rambuf, buf + pos, chunklen);

                /* send the chunk */
                if ((err = wlp_send(sock, rambuf, chunklen)) < 0) {
                        free(rambuf);
                        return err;
                }

                /* update input buffer position by the number of bytes
                 * actually sent.
                 */
                pos += err;
        }
        free(rambuf);
        return 0;
}

/* Convenience function for writint a NULL terminated string to the socket */
static int reply_str_P(int sock, const char *str)
{
        return reply_buf_P(sock, str, strlen_P(str));
} 

#endif /* __avr8__ */


static int reply_fmt(int sock, const char *format, ...)
{
	int ret;
	va_list ap;
        char *str;
        
        if ((str = malloc(HTTP_BUFSIZE)) < 0)
                return -1;

	va_start(ap, format);
	if ((ret = vsniprintf(str, HTTP_BUFSIZE, format, ap)) >= HTTP_BUFSIZE) {
                free(str);
                return -1;
        }

        reply_str(sock, str);
        free(str);
	va_end(ap);
	return ret;
}
