/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILIT76Y, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG
#define OWL_FILE "posixfs"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_list.h>

#include <linux/util/posixfs.h>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <ftw.h>

struct posixfs {
        const char *nativedir;
        struct list_head openlist; /* open files */

        struct fs fs;
};

struct posixfs_file {
        FILE *f;
        int dirty;
        char name[16];
        struct list_head list;

        struct file file;
};

static struct file *posixfs_open(struct fs *fs, const char *name, int flags,
                                 uint32_t size);
static void posixfs_close(struct file *file);
static int posixfs_seek(struct file *file, int pos, int whence);
static int posixfs_read(struct file *file, char *buf, int len);
static int posixfs_write(struct file *file, const char *buf, int len);
static int posixfs_eof(struct file *file);
static int posixfs_format(struct fs *fs);
static int posixfs_rename(struct fs *fs, const char *old, const char *new);
static int posixfs_remove(struct fs *fs, const char *name);
static int posixfs_stat(struct fs *fs, const char *name, struct fstat *out);
static const char *posixfs_next(struct fs *fs, const char *name);
static struct ftl *posixfs_ftl(struct fs *fs);
static int posixfs_check(struct fs *fs, const char *name, int action);
static int unlink_cb(const char *fpath, const struct stat *sb, int typeflag);

struct fs *
posixfs_create(const char *dir)
{
        struct posixfs *priv;
        struct fs *fs;
        struct stat statbuf;
        
        owl_dbg("dir:%s", dir);

        if (owl_os_strlen(dir) > 15)
                return NULL;

        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL) {
                owl_err("fail to alloc priv");
                return NULL;
        }

        if (stat(dir, &statbuf) < 0) {
                owl_info("creating dir");
                if (mkdir(dir, 0700) < 0) {
                        owl_err("fail to create dir, err:%d", errno);
                        owl_os_free(priv);
                        return NULL;
                }
        }

        fs = &priv->fs;
        
        fs->open = posixfs_open;
        fs->close = posixfs_close;
        fs->seek = posixfs_seek;
        fs->read = posixfs_read;
        fs->write = posixfs_write;
        fs->eof = posixfs_eof;
        fs->format = posixfs_format;
        fs->rename = posixfs_rename;
        fs->remove = posixfs_remove;
        fs->stat = posixfs_stat;
        fs->next = posixfs_next;
        fs->check = posixfs_check;
        fs->ftl = NULL;

        priv->nativedir = dir;
        list_init(&priv->openlist);
        
        return &priv->fs;
}

void
posixfs_destroy(struct fs *fs)
{
        struct posixfs *priv;

        owl_dbg("fs:%p", fs);

        priv = container_of(fs, struct posixfs, fs);

        if (!list_empty(&priv->openlist)) {
                owl_err("warning all files not closed");
        }

        owl_os_free(priv);
}


static struct file *
posixfs_open(struct fs *fs, const char *name, int flags, uint32_t size)
{
        struct posixfs *priv;
        struct posixfs_file *pfile;
        char path[32];
        char mode[3];

        priv = container_of(fs, struct posixfs, fs);
        
        if (owl_os_strlen(name) > 15) {
                owl_err("filename too long");
                return NULL;
        }
        
        if ((pfile = owl_os_zalloc(sizeof(*pfile))) == NULL) {
                owl_err("could not alloc file");
                return NULL;
        }

        if (owl_bit(flags, FS_CREATE) && owl_bit(flags, FS_TRUNC))
                owl_os_strcpy(mode, "w");
        else if (owl_bit(flags, FS_CREATE))
                owl_os_strcpy(mode, "w+");
        else if (owl_bit(flags, FS_RDONLY))
                owl_os_strcpy(mode, "r");
        else
                owl_os_strcpy(mode, "r+");
        
        owl_os_strcpy(path, priv->nativedir);
        owl_os_strcat(path, "/");
        owl_os_strcat(path, name);
        
        pfile->f = fopen(path, mode);
        if (pfile->f == NULL) {
                owl_err("unable to open file");
                owl_os_free(pfile);
                return NULL;
        }

        owl_os_strncpy(pfile->name, name, sizeof(pfile->name) - 1);
        list_add_tail(&pfile->list, &priv->openlist);
        return &pfile->file;
}


static void
posixfs_close(struct file *file)
{
        struct posixfs_file *pfile;
        
        owl_dbg("file:%p", file);

        pfile = container_of(file, struct posixfs_file, file);
        
        list_del(&pfile->list);
        fclose(pfile->f);
        owl_os_free(pfile);
}


static int
posixfs_seek(struct file *file, int pos, int whence)
{
        struct posixfs_file *pfile;
        int xwhence;
        
        owl_dbg("file:%p pos:%d whence:%d", file, pos, whence);
        
        pfile = container_of(file, struct posixfs_file, file);

        switch (whence) {
        case FS_SEEK_SET:
                xwhence = SEEK_SET;
                break;
        case FS_SEEK_END:
                xwhence = SEEK_END;
                break;
        case FS_SEEK_CUR:
                xwhence = SEEK_CUR;
                break;
        };
        
        if (fseek(pfile->f, pos, xwhence) < 0) {
                owl_err("failed to seek");
                return OWL_FAILURE;
        }

        return 0;
}

static int
posixfs_read(struct file *file, char *buf, int len)
{
        int err;
        struct posixfs_file *pfile;
        
        owl_vdbg("file:%p buf:%p len:%d", file, buf, len);
        
        pfile = container_of(file, struct posixfs_file, file);
        
        if ((err = fread(buf, 1, len, pfile->f)) < 0) {
                owl_err("fail to read err:%d len:%d", err, len);
                return OWL_FAILURE;
        }

        return err;
}

static int
posixfs_write(struct file *file, const char *buf, int len)
{
        int err;
        struct posixfs_file *pfile;
        
        owl_vdbg("file:%p buf:%p len:%d", file, buf, len);

        pfile = container_of(file, struct posixfs_file, file);
        
        if ((err = fwrite(buf, 1, len, pfile->f)) < 0) {
                owl_err("fail to write err:%d len:%d", err, len);
                return OWL_FAILURE;
        }

        pfile->dirty = 1;
        return err;
}

static int
posixfs_eof(struct file *file)
{
        struct posixfs_file *pfile;

        owl_vdbg("file:%p", file);
        
        pfile = container_of(file, struct posixfs_file, file);

        if (feof(pfile->f))
                return 1;
        else
                return 0;
}

static int
posixfs_format(struct fs *fs)
{
        struct posixfs *priv;
        int err;
        
        owl_dbg("fs:%p", fs);

        priv = container_of(fs, struct posixfs, fs);
        
        if ((err = ftw(priv->nativedir, unlink_cb, 64)) < 0) {
                owl_err("nftw failed, err:%d", err);
        }

        return err;
}


static int
posixfs_rename(struct fs *fs, const char *old, const char *new)
{        
        struct posixfs *priv;
        char oldpath[32];
        char newpath[32];
        int err;

        owl_dbg("fs:%p old:%s new:%s", fs, old, new);

        priv = container_of(fs, struct posixfs, fs);
        owl_os_snprintf(
                oldpath, sizeof(oldpath), "%s/%s", priv->nativedir, old);
        owl_os_snprintf(
                newpath, sizeof(newpath), "%s/%s", priv->nativedir, new);
        
        if ((err = rename(oldpath, newpath)) < 0) {
                owl_err("failed to rename, err:%d", err);
                return OWL_FAILURE;
        }

        return 0;
}


static int
posixfs_remove(struct fs *fs, const char *name)
{
        struct posixfs *priv;
        int err;
        char path[32];
        
        owl_dbg("fs:%p name:%s", fs, name);

        priv = container_of(fs, struct posixfs, fs); 
        owl_os_snprintf(path, sizeof(path), "%s/%s", priv->nativedir, name);

        if ((err = remove(path)) < 0) {
                owl_err("fail to remove, err:%d", err);
                return OWL_FAILURE;
        }

        return 0;
}


static int
posixfs_stat(struct fs *fs, const char *name, struct fstat *out)
{
        struct posixfs *priv;
        struct stat statbuf;
        int err;
        char path[32];
        struct list_head *pos = NULL;
                
        owl_vdbg("fs:%p name:%s stat:%p", fs, name, stat);

        priv = container_of(fs, struct posixfs, fs); 
        owl_os_snprintf(path, sizeof(path), "%s/%s", priv->nativedir, name);

        if ((err = stat(path, &statbuf)) < 0) {
                owl_err("stat failed, err:%d", errno);
                return OWL_FAILURE;
        }

        out->size = statbuf.st_size;
        out->capacity = 0;
        out->open = 0;
        out->flags = 0;
        out->ftl_addr = 0;
        out->file = NULL;
        
        list_for_each(pos, &priv->openlist) {
                struct posixfs_file *iter =
                        list_entry(pos, struct posixfs_file, list);

                if (owl_os_strcmp(iter->name, name))
                        continue;
                
                out->open++;
                if (iter->dirty)
                        out->flags |= FS_DIRTY;
                
                out->file = &iter->file;
                
        }
        
        return 0;
}


static const char *
posixfs_next(struct fs *fs, const char *name)
{
        struct posixfs *priv;
        DIR *dir;
        struct dirent *entry;
        int next = 0;
        static char filename[FILENAME_MAX];

        filename[0] = 0;
        priv = container_of(fs, struct posixfs, fs);
        
        if ((dir = opendir(priv->nativedir)) == NULL) {
                owl_err("could not open dir, err:%d", errno);
                return NULL;
        }
        
        while (entry = readdir(dir)) {

                if (entry->d_type != DT_REG)
                        continue;
                
                if (name == NULL || next) {
                        owl_os_strcpy(filename, entry->d_name);
                        owl_info("filename:%s", filename);
                        break;
                }
                
                if (owl_os_strcmp(name, entry->d_name))
                        continue;

                next = 1;
        }

        closedir(dir);

        if (owl_os_strlen(filename))
                return filename;
        else
                return NULL;
}


static int
posixfs_check(struct fs *fs, const char *name, int action)
{        
        owl_dbg("fs:%p name:%s action:%d", fs, name, action);
        return 0;
}


static int
unlink_cb(const char *fpath, const struct stat *sb, int typeflag)
{
        int err;

        owl_dbg("fpath:%s typeflag:%d", fpath, typeflag);

        if ((err = remove(fpath)) < 0) {
                owl_err("fail to remove fpath:%d", err);
        }

        return err;
}
