/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO

#include <avr/io.h>
#include <util/delay_basic.h>

#define F_CPU 32000000UL
#include <util/delay.h>
#include "ioport.h"
#include "board.h"

#include <board.h>
#include <sysclk.h>

#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_uart.h>

int
owluart_init(void *uart,
             uint32_t baudrate, enum uart_databits databits,
             enum uart_parity parity, enum uart_stopbits stopbits,
             int rtscts)
{
        USART_t *port = uart;
        uint8_t ctrlc = 0;
        uint8_t ctrlb = 0;
        uint16_t bsel;

#if BOARD == XPLAIN
        /* usb-to-serial on xplain seems broken for other configurations */
        if (port == &USARTC0) {
                if (baudrate != 9600 ||
                    databits != 8 ||
                    stopbits != 1 ||
                    parity != UART_PARITY_NONE ||
                    rtscts)
                        return OWL_ERR_NOTSUPP;
        }
#endif
        
        if (rtscts)
                return OWL_ERR_NOTSUPP;

        switch(databits) {
        case UART_DATABITS_5:
                ctrlc |= USART_CHSIZE_5BIT_gc;
                break;
        case UART_DATABITS_6:
                ctrlc |= USART_CHSIZE_6BIT_gc;
                break;
        case UART_DATABITS_7:
                ctrlc |= USART_CHSIZE_7BIT_gc;
                break;
        case UART_DATABITS_8:
                ctrlc |= USART_CHSIZE_8BIT_gc;
                break;
        case UART_DATABITS_9:
                ctrlc |= USART_CHSIZE_9BIT_gc;
                break;
        };
            
        switch (parity) {
        case UART_PARITY_NONE:
                ctrlc |= USART_PMODE_DISABLED_gc;
                break;
        case UART_PARITY_EVEN:
                ctrlc |= USART_PMODE_EVEN_gc;
                break;
        case UART_PARITY_ODD:
                ctrlc |= USART_PMODE_ODD_gc;
                break;
        default:
                return OWL_ERR_NOTSUPP;
        }
        
        switch (stopbits) {
        case UART_STOPBITS_1:
                break;
        case UART_STOPBITS_2:
                ctrlc |= USART_SBMODE_bm;
                break;
        };
 
        ctrlb = USART_TXEN_bm | USART_RXEN_bm;
        
        if (sysclk_get_per_hz() < (16UL * baudrate)) {
                /* need double transmission speed */
                ctrlb |= USART_CLK2X_bm;
                bsel = sysclk_get_per_hz() / (16UL * baudrate) - 1;
                
        } else {
                bsel = sysclk_get_per_hz() / (16UL * baudrate) - 1;
        }

        port->CTRLB = 0; /* disable transceiver while confiuring */
        port->CTRLC = ctrlc;
        port->BAUDCTRLA = bsel & 0xff;
        port->BAUDCTRLB = bsel >> 8;
        port->CTRLB = ctrlb;
        return 0;
}

void
owluart_outb(void *uart, int c)
{
        USART_t *port = uart;
        while (!(port->STATUS & USART_DREIF_bm));
        port->DATA = c;
        
#if BOARD == XPLAIN
        /* usb-to-serial hacks */
        if (port == &USARTC0) {
                while (!(port->STATUS & USART_TXCIF_bm));
                port->STATUS |= USART_TXCIF_bm;
                _delay_loop_2(1000);
        }
#endif
}

int
owluart_inb(void *uart, unsigned int timeout)
{
        unsigned int ms = owltmr_get_ms();
        USART_t *port = uart;

        if (timeout > 0)
                return OWL_ERR_NOTSUPP;
        
        for (;;) {
                if (owluart_poll(port)) {
                        int i = port->DATA;
                        owl_dbg("0x%x %c", i, (char) i); 
                        return i;
                }
                
                if (timeout > 0 &&
                    owltmr_intvl_passed(ms, owltmr_get_ms(), timeout)) {
                        owl_info("got timeout (%d)", timeout);
                        return OWL_ERR_TIMEOUT;
                }
        }
}

int
owluart_read(void *uart, void *buf, int len)
{
        int i;
        USART_t *port = uart;
        char *in = buf;

        for (i = 0; i < len; i++) {
                while (!(port->STATUS & USART_RXCIF_bm));
                if (buf == NULL)  /* support NULL buffer in e.g. wlp_recv() */
                        (void) port->DATA;
                else
                        in[i] = port->DATA;
        }

        return len;
}

int
owluart_write(void *uart, const void *buf, int len)
{
        int i;
        const char *out = buf;

        for (i = 0; i < len; i++)
                owluart_outb(uart, out[i]);
        
        return len;
}

int
owluart_poll(void *uart)
{
        USART_t *port = uart;
        owl_dbg("port:%x STATUS:%x CTRLA:%x CTRLB:%x CTRLC:%x",
                port, port->STATUS, port->CTRLA, port->CTRLB, port->CTRLC);
        return port->STATUS & USART_RXCIF_bm;
}

const char *
owluart_parity(enum uart_parity parity)
{
        switch (parity) {
        case UART_PARITY_NONE:
                return "N";
        case UART_PARITY_EVEN:
                return "E";
        case UART_PARITY_ODD:
                return "O";
        case UART_PARITY_MARK:
                return "M";
        case UART_PARITY_SPACE:
                return "S";
        };

        owl_assert(0);
        owl_err("invalid parity:%d", parity);
        return "";
}
