/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_MINIMAL
#define OWL_FILE "timer"

#include <owl/core/owl_timer.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_os.h>

struct owltmr_ctx {
        struct list_head timeouts;
        struct list_head *iter;
        volatile unsigned int cur_ts; /* current timestamp in ms */
#define OWLTMR_FLAG_POLL       (1 << 0)
#define OWLTMR_FLAG_FIRST_POLL (1 << 1) /* First time owltmr_poll() is called,
                                           the suspend_cb() should be invoked
                                           even though no timer was expired. */
#define OWLTMR_FLAG_SUSPEND_DISABLED (1 << 2)
        int flags;

        void (*suspend_cb)(void *ctx, uint32_t ms);
        void *suspend_ctx;

#if OWL_CONFIG_TIMER_STATS
        uint32_t poll_max; /* maximum time for owltmr_poll() */
        uint32_t cb_max; /* maximum time (ms) a single callback has been exec */
#endif
};


#if defined(OWL_DEBUG)
# define tmo_name(tmo) (tmo)->name
#else
# define tmo_name(tmo) ""
#endif

static void suspend(void);
static void list_del_safe(struct list_head *entry, struct list_head **n);

static struct owltmr_ctx *priv = NULL;

int
owltmr_init(void)
{
        priv = owl_os_zalloc(sizeof(struct owltmr_ctx));
        if (priv == NULL) {
                owl_err("could not alloc priv");
                return OWL_ERR_MEM;
        }
        list_init(&priv->timeouts);
        owl_bitset(priv->flags, OWLTMR_FLAG_FIRST_POLL);
        return 0;
}

struct owltmr_tmo *
owltmr_alloc(void)
{
        struct owltmr_tmo *tmo = owl_os_zalloc(sizeof(struct owltmr_tmo));
        if (tmo == NULL) {
                owl_err("could not alloc tmo");
                return NULL;
        }
        owl_dbg("tmo:%p", tmo);
        return tmo;
}

void
owltmr_free(struct owltmr_tmo *tmo)
{
        if (tmo == NULL)
                return;

        owl_dbg("tmo:%p name:%s %s%s%s", tmo, tmo_name(tmo),
                (tmo->flags & TMO_PERIODIC) ? "PERIODIC " : "",
                (tmo->flags & TMO_SCHEDULED) ? "SCHEDULED " : "",
                (tmo->flags & TMO_CANCEL) ? "CANCEL " : "");
        
        if (owl_bit(tmo->flags, TMO_SCHEDULED))
                list_del_safe(&tmo->list, &priv->iter);

        owl_os_free(tmo);
}

void
owltmr_sched(struct owltmr_tmo *tmo, int ms, int periodic,
             owltmr_tmo_cb_t *cb, void *ctx)
{

        owl_info("tmo:%p name:%s ms:%d periodic:%d cb:%p ctx:%p",
                 tmo, tmo_name(tmo), ms, periodic, cb, ctx);
        
        owl_bitclr(tmo->flags, TMO_CANCEL);
        
        tmo->cb = cb;
        tmo->ctx = ctx;
        tmo->ts = priv->cur_ts;
        tmo->intvl = ms;
        if (periodic)
                owl_bitset(tmo->flags, TMO_PERIODIC);

        /* First time add the timer to the list */
        
        if (!owl_bit(tmo->flags, TMO_SCHEDULED))
                list_add_tail(&tmo->list, &priv->timeouts);

        owl_bitset(tmo->flags, TMO_SCHEDULED);
        owl_vdbg("done");
}

void
owltmr_cancel(struct owltmr_tmo *tmo)
{
        if (!owl_bit(tmo->flags, TMO_SCHEDULED))
                return;

        owl_dbg("cancel tmo:%p name:%s", tmo, tmo_name(tmo));
        list_del_safe(&tmo->list, &priv->iter);
        owl_bitclr(tmo->flags, TMO_SCHEDULED);
        owl_vdbg("done");
}

bool_t
owltmr_is_sched(struct owltmr_tmo *tmo)
{
        return owl_bit(tmo->flags, TMO_SCHEDULED);
}

struct owltmr_tmo *
owltmr_addtmo(int ms, int periodic, owltmr_tmo_cb_t *cb, void *ctx)
{
        struct owltmr_tmo *tmo;
        
        if ((tmo = owltmr_alloc()) == NULL) {
                owl_err("could not alloc tmo");
                return NULL;
        }

        owltmr_sched(tmo, ms, periodic, cb, ctx);
        return tmo;
}

void
owltmr_deltmo(struct owltmr_tmo *tmo)
{
        owltmr_free(tmo);
}

void
owltmr_modtmo(struct owltmr_tmo *tmo, int ms)
{
        owltmr_sched(
                tmo, ms, owl_bit(tmo->flags, TMO_PERIODIC),
                tmo->cb, tmo->ctx);
}

unsigned int
owltmr_get_ms(void)
{
        return priv->cur_ts;
}

void
owltmr_tick(unsigned int inc)
{
        if (priv == NULL)
                return;
        priv->cur_ts += inc;
}

void
owltmr_mdelay(int ms)
{
        unsigned int old = priv->cur_ts;
        owl_dbg("busy waiting until:%d now:%d [ms]",
                old + ms, priv->cur_ts);
        while (!owltmr_intvl_passed(old, priv->cur_ts, ms));
}

void
owltmr_poll(void)
{
        struct list_head *pos;

        bool_t event = 0;

#if OWL_CONFIG_TIMER_STATS
        uint32_t poll_start, poll_time;
        poll_start = priv->cur_ts;
#endif /* OWL_CONFIG_TIMER_STATS */

        owl_vdbg("enter");

        /* Always invoke suspend_cb first time poll is called since there might
           be a while until the first timer expires. */
        if (owl_bit(priv->flags, OWLTMR_FLAG_FIRST_POLL)) {
                event = 1;
                owl_bitclr(priv->flags, OWLTMR_FLAG_FIRST_POLL);
        }

        owl_bitset(priv->flags, OWLTMR_FLAG_POLL); 

        list_for_each_safe(pos, priv->iter, &priv->timeouts) {
                struct owltmr_tmo *iter;
                iter = list_entry(pos, struct owltmr_tmo, list);
                
                if (!owltmr_intvl_passed(iter->ts, priv->cur_ts, iter->intvl))
                        continue;

                owl_vdbg("invoking tmo:%p name:%s", iter, tmo_name(iter));

                /* set flag before invoking cb() since owltmr_modtmo() might
                 * be invoked from the cb
                 */
                if (owl_bit(iter->flags, TMO_PERIODIC)) {
                        iter->ts = priv->cur_ts;
                } else {

                        /* Remove oneshot timers _before_ invoking the callback.
                           This allows us to free a one-show timer from its
                           callback. For periodic timers, the user have to
                           invoke owltmr_cancel() first. */
                        list_del(&iter->list);
                        owl_bitclr(iter->flags, TMO_SCHEDULED);
                }
                
                if (iter->cb) {
#if OWL_CONFIG_TIMER_STATS
                        uint32_t cb_start, cb_time;
                        cb_start = priv->cur_ts;
#endif /* OWL_CONFIG_TIMER_STATS */
                        iter->cb(iter, iter->ctx);
#if OWL_CONFIG_TIMER_STATS
                        cb_time = owltmr_get_intvl(cb_start, priv->cur_ts);
                        if (cb_time > priv->cb_max)
                                priv->cb_max = cb_time;
#endif /* OWL_CONFIG_TIMER_STATS */
                }
                event = 1; /* a timer expired, so we should call the wakeup_cb
                               to indicate that the time which next timeout
                               occurs might be changed */
        }

        owl_bitclr(priv->flags, OWLTMR_FLAG_POLL);

        priv->iter = NULL;

#if OWL_CONFIG_TIMER_STATS
        poll_time = owltmr_get_intvl(poll_start, priv->cur_ts);
        if (poll_time > priv->poll_max)
                priv->poll_max = poll_time;
#endif /* OWL_CONFIG_TIMER_STATS */
        
        owl_vdbg("exit event:%d", event);

        if (!owl_bit(priv->flags, OWLTMR_FLAG_SUSPEND_DISABLED) &&
            (event || list_empty(&priv->timeouts)))
                suspend();

}

int
owltmr_intvl_passed(unsigned int old, unsigned int new, unsigned int diff)
{
        if (diff && old == new)
                return 0;
        else if (owltmr_get_intvl(old, new) >= diff)
                return 1;
        else
                return 0;
}

unsigned int
owltmr_get_intvl(unsigned int old, unsigned int new)
{
        /* wrap */
        if (new < old)
                return ((unsigned int) (-1) - old) + new;
        
        /* no wrap */
        return new - old;
}

void
owltmr_set_suspend(int enable)
{
        if (enable)
                owl_bitclr(priv->flags, OWLTMR_FLAG_SUSPEND_DISABLED);
        else
                owl_bitset(priv->flags, OWLTMR_FLAG_SUSPEND_DISABLED);
}

void
owltmr_set_suspend_cb(void suspend_cb(void *ctx, uint32_t ms), void *ctx)
{
        priv->suspend_cb = suspend_cb;
        priv->suspend_ctx = ctx;
}

#if defined(OWL_DEBUG)
void
owltmr_set_name(struct owltmr_tmo *tmo, const char *name)
{
        tmo->name = name;
}
#endif

static void
suspend(void)
{
        struct list_head *pos;
        uint32_t next = 0;
        uint32_t ts = priv->cur_ts;
        
        list_for_each(pos, &priv->timeouts) {
                uint32_t intvl;
                struct owltmr_tmo *iter;
                iter = list_entry(pos, struct owltmr_tmo, list);
                
                if (!owl_bit(iter->flags, TMO_SCHEDULED))
                        continue;
                if (owl_bit(iter->flags, TMO_CANCEL))
                        continue;

                /* A timer has already passed its deadline, so we should
                   not sleep. This might occur e.g. if we spend too long time
                   in a callback */
                if (owltmr_intvl_passed(iter->ts, ts, iter->intvl))
                        return;
                
                intvl = owltmr_get_intvl(ts, iter->ts + iter->intvl);

                owl_vdbg("tmo:%p name:%s intvl:%d expire at:%u in:%u",
                         iter, tmo_name(iter), iter->intvl,
                         iter->ts + iter->intvl, intvl);

                if (!next || intvl < next)
                        next = intvl;
        }

        owl_vdbg("ts:%u, next timeout in %u ms", ts, next);
        if (priv->suspend_cb)
                priv->suspend_cb(priv->suspend_ctx, next);
}

#if OWL_CONFIG_TIMER_STATS
static void
timer_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        if (argc > 1) {
                if (!owl_os_strcmp("reset", argv[1])) {
                        priv->poll_max = 0;
                        priv->cb_max = 0;
                        
                } else {
                        goto usage;
                }
                
        } else {

                owlsh_printf(sh, "Poll:      %lu\r\n", priv->poll_max);
                owlsh_printf(sh, "Callback:  %lu\r\n", priv->cb_max);
                owlsh_printf(sh, "\n\r");
        }

        return;

usage:
        owlsh_printf(sh, "usage: timer [reset]\n\r");
}

const struct owlsh_cmd cmd_timer = {
        "timer", timer_f, "timer stats"
};
#endif /* OWL_CONFIG_TIMER_STATS */


/**
 * list_del_safe - deletes entry from list.
 * @entry: the element to delete from the list.
 * @n: the next iterator position (parameter @n in list_for_each_safe).
 *
 * Note: Safe to call to delete another list entry while iterating with
 * list_for_each_safe(). This function is needed to handle the case where
 * a timer is deleted while we are iterating through the timer list.
 * list_for_each_safe() only supports removing the current position, not e.g.
 * the next position. list_del_safe() handles this for us.
 */
static void
list_del_safe(struct list_head *entry, struct list_head **n)
{
        if (*n == entry)
                *n = (*n)->next;

        list_del(entry);
}
