/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_MINIMAL
#define OWL_FILE "wlp_cloud"
#include <owl/core/owl_debug.h>
#include <owl/core/owl_list.h>
#include <owl/core/owl_err.h>

#include <pico/wlp_cloud.h>
#include <pico/wlp_netlist.h>

#include <wlp_api.h>

struct scan_ctx {
#define FLAG_FREE_NETLIST (1 << 0)
        int flags;
        int idx;
        struct netlist *netlist;
};

static int scan_sent_cb(struct cloud_req *cloud_req);
static int scan_recv_cb(struct cloud_req *cloud_req);
static void scan_err_cb(struct cloud_req *cloud_req, int err);
static void netlist_done_cb(void *ctx, struct netlist *netlist);

static int
scan_sent_cb(struct cloud_req *cloud_req)
{
        int err;
        struct scan_ctx *priv = cloud_req_priv(cloud_req);
        
        owl_dbg("idx:%d", priv->idx);

        if (priv->idx < 0) {
                struct cloud_scan_cfm cfm;
                cfm.count = owl_htonl(priv->netlist->count);
                if ((err = cloud_send(cloud_req, &cfm, sizeof(cfm), 0))) {
                        owl_info("could not send cfm, err:%d", err);
                        return err;
                }

                priv->idx++;
        }
        
        for (; priv->idx < (int) priv->netlist->count; priv->idx++) {
                struct cloud_net dstnet;
                struct wlp_network *srcnet = &priv->netlist->net[priv->idx];
                
                /* convert to cloud_net */
                dstnet.ssid_len = srcnet->ssid.len;
                owl_os_memcpy(dstnet.ssid, srcnet->ssid.ssid,
                              sizeof(dstnet.ssid));
                owl_os_memcpy(dstnet.bssid, srcnet->bssid.octet,
                              sizeof(dstnet.bssid));
                dstnet.auth = srcnet->auth;
                dstnet.channel = srcnet->channel;
                dstnet.type = srcnet->type;
                dstnet.rssi = owl_htons(srcnet->rssi);
                dstnet.snr = owl_htons(srcnet->snr);
                        
                if ((err = cloud_send(cloud_req, &dstnet, sizeof(dstnet), 0))) {
                        owl_info("not enough room, err:%d", err);
                        return err;
                }
        }

        /* done */
        if (owl_bit(priv->flags, FLAG_FREE_NETLIST))
                owl_os_free(priv->netlist);
        owl_os_free(priv);
        cloud_req_set_priv(cloud_req, NULL);
        owl_info("done");
        return 0;
}

static int
scan_recv_cb(struct cloud_req *cloud_req)
{
        int err;
        struct scan_ctx *priv = cloud_req_priv(cloud_req);
        
        owl_dbg("enter");
        
        if (priv == NULL) {
                struct netlist *netlist = cloud_req_ctx(cloud_req);

                owl_info("netlist:%p", netlist);

                if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                        return OWL_ERR_MEM;

                priv->idx = -1;
                
                /* Create a netlist if not passed to cloud_addcmd() */
                if (netlist == NULL) {
                        owl_bitset(priv->flags, FLAG_FREE_NETLIST);
                        if ((err = netlist_create(
                                     NULL, 1, 20, netlist_done_cb, priv))) {
                                owl_err("netlist create failed");
                                owl_os_free(priv);
                                return err;
                        }
                        
                } else {
                        priv->netlist = netlist;
                }

                cloud_req_set_priv(cloud_req, priv);
        }
        
        /* netlist_done_cb not called yet */
        if (priv->netlist == NULL)
                return OWL_ERR_AGAIN;

        /* Scan is complete */
        owl_dbg("scan completed");
        
        /* response will be sent in sent_cb */
        cloud_req_set_sent(cloud_req, scan_sent_cb);
        cloud_req_set_err(cloud_req, scan_err_cb);
        cloud_req_set_cfmlen(
                cloud_req, sizeof(struct cloud_scan_cfm) +
                sizeof(struct cloud_net) * priv->netlist->count);

        return 0;
}

static void
scan_err_cb(struct cloud_req *cloud_req, int err)
{
        struct scan_ctx *priv = cloud_req_priv(cloud_req);
        owl_unused(err);
        owl_err("err:%d", err);
        
        if (priv) {
                if (owl_bit(priv->flags, FLAG_FREE_NETLIST))
                        owl_os_free(priv->netlist);
                
                owl_os_free(priv);
        }
}


static void
netlist_done_cb(void *ctx, struct netlist *netlist)
{
        struct scan_ctx *priv = ctx;
        owl_info("count:%d", netlist->count);
        priv->netlist = netlist;
}

const struct cloud_cmd cloud_scan = { CLOUD_ID_SCAN, scan_recv_cb };
