/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO
#define OWL_FILE "wlp_netlist"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>

#include <pico/wlp_netlist.h>

#include <wlp_api.h>


struct netlist_ctx {
        struct wlp_ssid ssid;
        uint8_t try;
        uint8_t maxcount;
        struct netlist *netlist;
        void (*done_cb)(void *ctx, struct netlist *netlist);
        void *ctx;
};

static void netlist_cb(void *ctx, struct wlp_network *net);
static void netlist_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void netlist_done_cb(void *ctx, struct netlist *netlist);

int
netlist_create(struct wlp_ssid *ssid, int tries, int maxcount, 
               void (*done_cb)(void *ctx, struct netlist *netlist), void *ctx)
{
        int err;
        struct netlist_ctx *priv;
        
        owl_info("maxcount:%d tries:%d", maxcount, tries);

        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL) {
                owl_err("could not alloc priv");
                return OWL_ERR_MEM;
        }
                
        priv->maxcount = maxcount;
        if ((priv->netlist = owl_os_zalloc(
                     sizeof(*priv->netlist) +
                     sizeof(struct wlp_network) * priv->maxcount)) == NULL) {
                owl_err("could not alloc netlist");
                owl_os_free(priv);
                return OWL_ERR_MEM;
        }

        priv->try = tries - 1;
        if (ssid != NULL)
                owl_os_memcpy(&priv->ssid, ssid, sizeof(priv->ssid));
        
        if ((err = wlp_scan(&priv->ssid, netlist_cb, priv))) {
                owl_err("fail to scan, err:%d", err);
                owl_os_free(priv->netlist);
                owl_os_free(priv);
                return err;
        }
        priv->done_cb = done_cb;
        priv->ctx = ctx;
        return 0;
}

static int
netlist_find(struct netlist_ctx *priv, uint8_t *bssid)
{
        unsigned int i;
        for (i = 0; i < priv->netlist->count; i++) {
                if (!owl_os_memcmp(
                            priv->netlist->net[i].bssid.octet, bssid,
                            sizeof(priv->netlist->net[i].bssid.octet)))
                        return i;
        }

        return -1;
}

/* scan result */
static void
netlist_cb(void *ctx, struct wlp_network *net)
{
        struct netlist_ctx *priv = ctx;
        struct netlist *netlist = priv->netlist;
        int err;
        int idx;
        
        owl_dbg("net:%p", net);
        
        if (net == NULL) {
                
                if (priv->try > 0) {
                        owl_dbg("scan, tries left:%d", priv->try);
                        if ((err = wlp_scan(&priv->ssid, netlist_cb, priv))) {
                                owl_err("fail to scan, err:%d", err);
                                if (priv->done_cb)
                                        priv->done_cb(priv->ctx, netlist);
                                owl_os_free(priv);
                                return;
                        }
                        priv->try--;
                        
                } else {
                        owl_info("done, cnt:%d", netlist->count);
                        if (priv->done_cb)
                                priv->done_cb(priv->ctx, netlist);
                        owl_os_free(priv);
                }
                
                return;
        }

        /* fill in network information */

        if ((idx = netlist_find(priv, net->bssid.octet)) < 0) {
                idx = netlist->count;
                
                if (netlist->count < priv->maxcount) {
                        netlist->count++;
                } else {
                        owl_info("too many nets");
                        return;
                }
        }

        owl_os_memcpy(&netlist->net[idx], net, sizeof(*net));
}



static void
netlist_done_cb(void *ctx, struct netlist *netlist)
{
	struct owlsh_ctx *sh = ctx;
	unsigned int i;

        if (owlsh_compact(sh)) {

#if OWL_CONFIG_SHELL_OUTPUT_COMPACT

                /* <ssid len>;<ssid>;<hwaddr>;
                   <none|wep|wpa|wpa-psk|wpa2|wpa2-psk>;<rssi> */

                for (i = 0; i < netlist->count; i++) {
                        struct wlp_network *net = &netlist->net[i];
                        char ssid_buf[WLP_SSID_MAX_LENGTH + 1];
                        const char *auth = "";
                        
                        owl_os_memset(ssid_buf, 0, sizeof(ssid_buf));
                        owl_os_memcpy(ssid_buf, net->ssid.ssid, net->ssid.len);

                        switch (net->auth) {
                        case WLP_AUTH_AUTO:
                                auth = "none";
                                break;
                        case WLP_AUTH_WEP:
                                auth = "wep";
                                break;
                        case WLP_AUTH_WPA_PSK:
                                auth = "wpa-psk";
                                break;
                        case WLP_AUTH_WPA2_PSK:
                                auth = "wpa2-psk";
                                break;
                        case WLP_AUTH_WPA:
                                auth = "wpa";
                                break;
                        case WLP_AUTH_WPA2:
                                auth = "wpa2";
                                break;
                        default:
                                owl_err("invalid auth:%d", auth);
                                owl_assert(0);
                        }

                        owlsh_printf(
                                sh,
                                "%d;%s;%02x:%02x:%02x:%02x:%02x:%02x;%s;%d\r\n",
                                net->ssid.len, ssid_buf,
                                net->bssid.octet[0], net->bssid.octet[1],
                                net->bssid.octet[2], net->bssid.octet[3],
                                net->bssid.octet[4], net->bssid.octet[5],
                                auth, net->rssi);
                        
                }
                owlsh_printf(sh, "\r\n");

#endif /* OWL_CONFIG_SHELL_OUTPUT_COMPACT */
        } else  {
                
#if OWL_CONFIG_SHELL_OUTPUT_FULL
                if (netlist->count == 0) {
                        owlsh_printf(sh, "no nets found\r\n");
                        owl_os_free(netlist);
                        return;
                }

                owlsh_printf(sh, "\r\n");
                owlsh_printf(sh,
                             "BSSID              TYPE   CH RSSI SNR  "
                             "AUTH      SSID\r\n");
                owlsh_printf(sh, "-----------------------------------"
                             "----------------------------------\r\n");
                for (i = 0; i < netlist->count; i++) {
                        struct wlp_network *net = &netlist->net[i];
                        char buf[WLP_SSID_MAX_LENGTH + 1];

                        /* bssid */
                        owlsh_printf(sh, "%02x:%02x:%02x:%02x:%02x:%02x  ",
                                     net->bssid.octet[0], net->bssid.octet[1],
                                     net->bssid.octet[2], net->bssid.octet[3],
                                     net->bssid.octet[4], net->bssid.octet[5]);

                        /* type, channel, rssi, snr */
                        owlsh_printf(sh, "%s  %02d %02d  ",
                                     net->type == WLP_CONN_TYPE_INFRA ?
                                     "INFRA" : "ADHOC",
                                     net->channel, (int) net->rssi);

                        /* snr */
                        if (net->snr != WLP_SNR_UNKNOWN)
                                owlsh_printf(sh, "%02d   ", net->snr);
                        else
                                owlsh_printf(sh, "N/A  ");

                        /* auth */
                        switch (net->auth) {
                        case WLP_AUTH_AUTO:
                                owlsh_printf(sh, "NONE      ");
                                break;
                        case WLP_AUTH_WEP:
                                owlsh_printf(sh, "WEP       ");
                                break;
                        case WLP_AUTH_WPA_PSK:
                                owlsh_printf(sh, "WPA-PSK   ");
                                break;
                        case WLP_AUTH_WPA2_PSK:
                                owlsh_printf(sh, "WPA2-PSK  ");
                                break;
                        case WLP_AUTH_WPA:
                                owlsh_printf(sh, "WPA       ");
                                break;
                        case WLP_AUTH_WPA2:
                                owlsh_printf(sh, "WPA2      ");
                                break;
                        }

                        /* ssid */
                        owl_os_memset(buf, 0, sizeof buf);
                        owl_os_memcpy(buf, net->ssid.ssid, net->ssid.len);
                        owlsh_printf(sh, "%s", buf);

                        owlsh_printf(sh, "\r\n");
                }
#endif /* OWL_CONFIG_SHELL_OUTPUT_FULL */
        }

        owl_os_free(netlist);
}

static void
netlist_f(struct owlsh_ctx *sh, int argc,  const char *argv[])
{
        int res;
        int tries = 1;
        int maxcount = 16;
        
        if (argc > 1)
                tries = owl_os_atoi(argv[1]);
        if (argc > 2)
                maxcount = owl_os_atoi(argv[2]);

        res = netlist_create(NULL, tries, maxcount, netlist_done_cb, sh);
        owlsh_perr(sh, res);
}

const struct owlsh_cmd cmd_netlist = { "netlist", netlist_f, "Network list" };
