/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


#define OWL_TRACE OWL_TRACE_INFO
#define OWL_FILE "wlp_thread"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_thread.h>

#include <owl/util/rmutex.h>

#include <wlp_api.h>
#include <wlx_api.h>

struct wlp_thread {
	struct rmutex lock;
};

static struct wlp_thread *priv = NULL;

#if defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_DBG
static const char *dbg_ssid(struct wlp_ssid *ssid);
#endif

int
wlx_init(int (*bus_f)(void *ctx, const void *tx, void *rx, int len),
	 void (*select_f)(void *ctx, int select),
	 void (*irqen_f)(void *ctx, int enable),
	 void (*firmware_f)(void *ctx, int state),
	 void *ctx,
	 int flags)
{
        int err;
	struct wlp_thread *wlp_thread;
	
	owl_dbg("flags:0x%08x", flags);

	if (priv != NULL) {
		owl_err("already initialized");
		return OWL_ERR_STATE;
	}

	/* Assign priv when all initialization is complete to avoid race */
	if ((wlp_thread = owl_os_zalloc(sizeof(*priv))) == NULL) {
		owl_err("could not alloc priv");
		return OWL_ERR_MEM;
	}
	
	if ((err = rmutex_init(&wlp_thread->lock))) {
		owl_err("fail to init rmutex, err:%d", err);
		owl_os_free(wlp_thread);
		return err;
	}

        if ((err = _wlx_init(
                     bus_f, select_f, irqen_f, firmware_f, ctx, flags))) {
                owl_err("wlx init failed, err:%d", err);
		rmutex_destroy(&wlp_thread->lock);
		owl_os_free(wlp_thread);
                return err;
        }

	priv = wlp_thread;
	return 0;	
}

void
wlx_shutdown(void)
{
	int err;
	struct wlp_thread *wlp_thread;

	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return;
        }
        
	_wlx_shutdown();

	/* Prevent other threads from entering wlp-functions by setting priv
	   to NULL _before_ free'ing the priv struct */
	wlp_thread = priv;
	priv = NULL;

	/* Any other thread waiting on rmutex_lock() will return with
	   OWL_ERR_DESTROYED */
	rmutex_destroy(&wlp_thread->lock);

	/* Now we are sure that no one will use the priv struct, so free it */
	owl_os_free(wlp_thread);
}


/* Called by interrupt handler */
void
wlp_irq(void)
{
	owl_vdbg("enter");
	
	_wlp_irq();
}


int
wlp_poll(void)
{
	int err;
	
	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	/* Note that any callback will be invoked while we hold the lock, but
	   this is ok since the mutex is recursive */
	err = _wlp_poll();

	rmutex_unlock(&priv->lock);
	return err;
}



int
wlp_scan(struct wlp_ssid *ssid,
         void (*cb)(void *ctx, struct wlp_network *net), void *ctx)
{
	int err;

	owl_dbg("ssid:%s", dbg_ssid(ssid));
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

        if ((err = _wlp_scan(ssid, cb, ctx))) {
                owl_err("fail to scan, err:%d", err);
        }

	rmutex_unlock(&priv->lock);
	return err;
}


void
wlp_set_link_cb(void (*cb)(void *ctx, int link), void *ctx)
{
	int err;
		
	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return;
	}
	
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return;
        }

	_wlp_set_link_cb(cb, ctx);
	rmutex_unlock(&priv->lock);
}


void
wlp_set_yield_cb(void (*cb)(void *ctx), void *ctx)
{
	int err;
		
	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return;
	}
	
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return;
        }

	_wlp_set_yield_cb(cb, ctx);
	rmutex_unlock(&priv->lock);
}


int
wlp_linkup(struct wlp_ssid *ssid, const char *key, enum wlp_auth_mode auth)
{
	int err;
	
	owl_dbg("ssid:%s key:%s auth:%d", dbg_ssid(ssid), key, auth);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_linkup(ssid, key, auth))) {
		owl_err("linkup failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}

int
wlp_linkdown(void)
{
	int err;
	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}
	
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_linkdown())) {
		owl_err("linkdown failed, err:%d", err);
	}
	
	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_set_wps(int enable)
{
	int err;
	owl_dbg("enable:%d", enable);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_set_wps(enable))) {
		owl_err("set_wps failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_get_hwaddr(struct wlp_mac_addr *hwaddr)
{
	int err;
	owl_dbg("hwaddr:%p", hwaddr);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}
		
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_get_hwaddr(hwaddr))) {
		owl_err("get_hwaddr failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_get_network(struct wlp_network *net)
{
	int err;
	owl_dbg("net:%p", net);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_get_network(net))) {
		owl_err("get_network failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_get_link_context(void (*cb)(void *ctx, void *linkctx, int len),
		     void *ctx)
{ 
	int err;
	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }
	
	if ((err = _wlp_get_link_context(cb, ctx))) {
		owl_err("get_link_context failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_set_link_context(void *linkctx)
{
	int err;
	owl_dbg("linkctx:%p", linkctx);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_set_link_context(linkctx))) {
		owl_err("set_link_context failed, err:%d", err);
	}
	
	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_set_mode(uint8_t mode)
{
	int err;
	owl_dbg("mode:%d", mode);
	
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_set_mode(mode))) {
		owl_err("set_mode failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_get_mode(uint8_t *mode)
{
	int err;
	owl_dbg("mode:%p", mode);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_get_mode(mode))) {
		owl_err("get_mode failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_set_channel(uint8_t channel,
		void (*cb)(void *ctx, int status), void *ctx)
{
	int err;
	owl_dbg("channel:%d", channel);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_set_channel(channel, cb, ctx))) {
		owl_err("set_channel failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_conf_ap(int hidden, int beacon_intvl, int inact_tmo,
	    const char *country)
{
	int err;
	owl_dbg("hidden:%d beacon_intvl:%d inact_tmo:%d country:%s",
		hidden, beacon_intvl, inact_tmo, country);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_conf_ap(hidden, beacon_intvl, inact_tmo, country))) {
		owl_err("conf_ap failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_set_ps(int enable)
{
	int err;
	owl_dbg("enable:%d", enable);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_set_ps(enable))) {
		owl_err("set_ps failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_conf_ps(uint8_t use_ps_poll, uint32_t ps_traffic_timeout,
            uint32_t ps_delay, uint8_t rx_all_dtim,
            uint16_t listen_interval)
{
	int err;
	owl_dbg("ues_ps_poll:%d ps_traffic_timeout:%d ps_delay:%d "
		"rx_all_dtim:%d listen_interval:%d",
		use_ps_poll, ps_traffic_timeout, ps_delay,
		rx_all_dtim, listen_interval);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_conf_ps(use_ps_poll, ps_traffic_timeout, ps_delay,
				rx_all_dtim, listen_interval))) {
		owl_err("conf_ps failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlp_get_fw_version(uint32_t *code, char *build, int build_len,
		   char *hw_rev, int hw_rev_len)
{
	int err;
	owl_dbg("code:%p build:%p build_len:%d hw_rev:%p hw_rev_len:%d",
		code, build, build_len, hw_rev, hw_rev_len);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_get_fw_version(code, build, build_len, hw_rev,
				       hw_rev_len))) {
		owl_err("wlp_get_fw_version failed, err:%d", err);
	}
	
	rmutex_unlock(&priv->lock);
	return err;
}


int wlp_get_fw_caps(uint8_t *cap_flags)
{
	int err;
	owl_dbg("cap_flags:%p", cap_flags);

	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_get_fw_caps(cap_flags))) {
		owl_err("wlp_get_fw_caps failed, err:%d", err);
	}

	rmutex_unlock(&priv->lock);
	return err;
}


#if !WLP_STACK_OFFLOAD
void
wlx_set_recv_cb(int (*cb)(void *ctx, int len), void *ctx)
{
	int err;
	owl_dbg("enter");

	if (priv == NULL) {
		owl_err("not initialized");
		return;
	}
	
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return;
        }

	_wlx_set_recv_cb(cb, ctx);

	rmutex_unlock(&priv->lock);
}


int
wlx_process_tx(void *ethhdr, int len, int totlen, void *wlxhdr)
{
	int err;
	owl_vdbg("ethhdr:%p len:%d totlen:%d wlxhdr:%p",
		 ethhdr, len, totlen, wlxhdr);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	err = _wlx_process_tx(ethhdr, len, totlen, wlxhdr);

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlx_tx(void *data, int len)
{
	int err;
	owl_vdbg("data:%p len:%d", data, len);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	err = _wlx_tx(data, len);

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlx_rx(void *data, int len)
{
	int err;
	owl_vdbg("data:%p len:%d", data, len);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}
	
	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	err = _wlx_rx(data, len);

	rmutex_unlock(&priv->lock);
	return err;
}


int
wlx_process_rx(void *data, int len, void **stripped, int *stripped_len)
{
	int err;
	owl_vdbg("data:%p len:%d stripped:%p stripped_len:%p",
		 data, len, stripped, stripped_len);
	
	if (priv == NULL) {
		owl_err("not initialized");
		return OWL_ERR_STATE;
	}

	if ((err = rmutex_lock(&priv->lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	err = _wlx_process_rx(data, len, stripped, stripped_len);

	rmutex_unlock(&priv->lock);
	return err;
}
#endif /* !WLP_STACK_OFFLOAD */


#if defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_DBG
static const char *
dbg_ssid(struct wlp_ssid *ssid)
{
        static char ssid_str[33];

	if (ssid == NULL)
		return "";
	
	owl_os_memcpy(ssid_str, ssid->ssid, ssid->len);
	ssid_str[ssid->len] = 0;
	return ssid_str;
}
#endif
