/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#include "wlp_api.h"
#include "wlp_proto.h"
#include "wlp_inet.h"

#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_board.h>

/* Note that wlp_api uses malloc() and free() to allocate dynamic memory. If the
   environment does not support dynamic memory allocation, then WLP_HEAP
   define should be defined to 0 */
#if !defined(WLP_HEAP)
# define WLP_HEAP 1
#endif

#if !WLP_HEAP
/* If WLP_HEAP is set to 0, a static number of sockets will be pre-allocated.
   The WLP_SOCK_NUM defined can be set to change the number of available
   sockets */
# if !defined(WLP_SOCK_NUM)
#  define WLP_SOCK_NUM 8
# endif
#endif /* WLP_HEAP */

/* If WLP_LIBC is set to 1, we will use the clib functions directly, otherwise,
   we will go through the owl os layer (wol_os.h). If you are using wlp_api
   without the owl toolkit, make sure to define WLP_LIBC to 1*/
#if !defined(WLP_LIBC)
# define WLP_LIBC 1
#endif

#if WLP_LIBC
# include <string.h>
# include <stdlib.h>
# define owl_os_zalloc(x) calloc(x, 1)
# define owl_os_free      free
# define owl_os_strcpy    strcpy
# define owl_os_strncpy   strncpy
# define owl_os_strlen    strlen
# define owl_os_memcpy    memcpy
# define owl_os_memset    memset
#else
# include <owl/core/owl_os.h>
#endif

/* debugging only possible if using owl toolkit code */
#if defined(WLP_DEBUG)
# define OWL_TRACE OWL_TRACE_DBG
# include <owl/core/owl_debug.h>
static int debug = 1; /* use to control debugging from wlp_poll() */
# define dbg_allow() debug = 1;
# define dbg_inhibit() debug = 0;
# define dbg_allowed(x) if (debug) { x; }
#else /* define empty macros below since we don't have owl_debug.h */
# define owl_vdbg(...)
# define owl_dbg(...)
# define owl_info(...)
# define owl_err(...)
# define owl_hexdump(...)
# define dbg_allow()
# define dbg_inhibit()
# define dbg_allowed(x)
#endif

#define IP_ADDR_ANY_VALUE 0x00000000UL
#define IP_ADDR_BROADCAST_VALUE 0xffffffffUL

/* used by IP_ADDR_ANY and IP_ADDR_BROADCAST in ip_addr.h */
const struct ip_addr ip_addr_any = { IP_ADDR_ANY_VALUE };
const struct ip_addr ip_addr_broadcast = { IP_ADDR_BROADCAST_VALUE };


struct wlp_sock_t {
        int16_t id;
#define WLP_SOCK_FLAG_CONNECTED    (1 << 0) /* last known state for cb's*/
#define WLP_SOCK_FLAG_ERROR        (1 << 1) /* last known state for cb's*/
#define WLP_SOCK_FLAG_DELETED      (1 << 2)
        uint8_t flags;
        struct wlp_sock_t *next;
        void (*listen)(void *ctx, int sockid);
        void (*conn)(void *ctx, int sockid, int connected);
        void (*rx)(void *ctx, int sockid, int len);
        void *ctx; /* callback context for listen and conn */
        void *rx_ctx; /* callback context for rx */
};

struct wlp_ctx {

        int (*read)(void *ctx, void *data, int len);
        int (*write)(void *ctx, const void *data, int len);
        void (*select)(void *ctx, int select);
        void *rw_ctx;

        void (*wakeup)(void *ctx, int enable);
        void *wakeup_ctx;

        void (*yield)(void *ctx);
        void *yield_ctx;

#define WLP_FLAG_LINKUP   (1 << 0)
#define WLP_FLAG_GOT_ADDR (1 << 1)
#define WLP_FLAG_POLL     (1 << 2)
#define WLP_FLAG_SPI      (1 << 3)
#define WLP_FLAG_FW_UPGRADE (1 << 4)
        uint8_t flags;

        volatile uint8_t irq; /* accessed from interrupt context */
        
        void (*link)(void *ctx, int link);
        void *link_ctx;

        void (*addr)(void *ctx, const struct ip_addr *ip);
        void *addr_ctx;

        void (*lookup)(void *ctx, const struct ip_addr *ip);
        void *lookup_ctx;

        void (*linkctx)(void *ctx, void *linkctx, int len);
        void *linkctx_ctx;

        void (*fw_up_status)(void *ctx, uint8_t success);
        void *fw_up_status_ctx;

        void (*scan)(void *ctx, struct wlp_network *net);
        void *scan_ctx;

        struct wlp_msg msg;
        
#if !WLP_HEAP
        struct wlp_sock_t sock[WLP_SOCK_NUM];
#endif /* !WLP_HEAP */

} wlp_ctx;

struct wlp_ctx *priv = NULL;
struct wlp_sock_t *sock_head = NULL;

/* wlp api helpers */
static uint8_t get_dev_status(void);
static int net_config(const struct ip_addr *ip, const struct ip_addr *netmask,
                      const struct ip_addr *gw, const struct ip_addr *dns,
                      int dhcp);
static int net_info(struct ip_addr *ip, struct ip_addr *netmask,
                    struct ip_addr *gw, struct ip_addr *dns, int *dhcp,
                    struct ip_addr *dns_lookup, int *dns_result);
static int sock_status(int16_t sockid, int8_t *state);

/* send/recv wlp_proto messages */
static int send_mgmt_req(struct wlp_msg *msg, int id, int reqlen, int cfmlen);
static int send_mgmt_req_ext(struct wlp_msg *msg, int id, int reqlen,
                             int cfmlen, int claimbus);
static int send_data_req(const char *data, int len);
static int recv_data_ind(char *data, int len);

/* socket management functions */
static struct wlp_sock_t *sock_alloc(void);
static struct wlp_sock_t *sock_find(int id);
static void sock_free(struct wlp_sock_t *sock);

/* bus access functions */
static int bus_read(void *data, int len);
static int bus_write(const void *data, int len);
static void bus_claim(void);
static void bus_release(void);
static void bus_wait(void);

int
_wlp_init(int interface,
          int (*readf)(void *ctx, void *data, int len),
          int (*writef)(void *ctx, const void *data, int len),
          void (*selectf)(void *ctx, int select),
          void *ctx)
{

        owl_dbg("interface:%d readf:%p writef:%p ctx:%p",
                interface, readf, writef, ctx);
        
        priv = &wlp_ctx;
        owl_os_memset(priv, 0, sizeof(*priv));

        priv->read = readf;
        priv->write = writef;
        priv->select = selectf;
        priv->rw_ctx = ctx;

        if (interface == WLP_INTERFACE_SPI) {
                priv->flags |= WLP_FLAG_SPI;
                owl_dbg("init complete");
                return 0;
        }
        
        owl_dbg("init complete");
        return 0;
}

int
_wlp_get_fw_caps(uint8_t *cap_flags)
{
        struct wlp_get_config_cfm *cfm = &priv->msg.u.cfm.u.fw_config;
        int err;
        
        if ((err = send_mgmt_req(&priv->msg, WLP_GET_CONFIG, 0,
                                 sizeof(*cfm))) < 0)
                return err;

        *cap_flags = cfm->flags;
        return 0;
}

int
_wlp_set_wakeup_cb(void (*wakeupf)(void *ctx, int enable),
                   void *ctx) {
        struct wlp_set_config_req *req = &priv->msg.u.req.u.config;
        int err;
        owl_dbg("register wakeup callback");

        if (wakeupf)
                req->flags = WLP_FLAG_WAKEUP_PIN_PRESENT;
        else
                req->flags = 0;

        err = send_mgmt_req(&priv->msg, WLP_SET_CONFIG, sizeof(*req), 0);
        /* Don't set the callback until after the set-config message
         * has been sent. This stops send_mgmt_req() from calling
         * the wakeup callback until the SPB800 has been properly
         * set up for it. */
        priv->wakeup = wakeupf;
        priv->wakeup_ctx = ctx;
        /* Allow the SPB800 to sleep */
        priv->wakeup(priv->wakeup_ctx, 0);

        return err;
}

void
_wlp_set_yield_cb(void (*yieldf)(void *ctx), void *ctx)
{
        owl_dbg("yieldf:%p ctx:%p", yieldf, ctx);
        
        priv->yield = yieldf;
        priv->yield_ctx = ctx;
}


void
_wlp_irq(void)
{
        owl_vdbg("enter");
        priv->irq = 1;
}


int
_wlp_set_baudrate(uint32_t baudrate, uint8_t rtscts)
{
        struct wlp_uart_config_req *req = &priv->msg.u.req.u.uart_config;
        
        owl_dbg("baudrate:%u rtscts:%d", baudrate, rtscts);

        req->baudrate = htonl(baudrate);
        req->rtscts = rtscts;

        return send_mgmt_req(&priv->msg, WLP_UART_CONFIG,
                        sizeof(*req), sizeof(struct wlp_uart_config_cfm));
}

int
_wlp_get_fw_version(uint32_t *code, char *build, int build_len, 
                    char *hw_rev, int hw_rev_len)
{
        struct wlp_fw_version_cfm *cfm = &priv->msg.u.cfm.u.fw_version;
        int err;

        owl_dbg("code:%p build:%p len:%d hw_rev:%s len:%d", code, build,
                build_len, hw_rev, hw_rev_len);
        
        if ((err = send_mgmt_req(&priv->msg, WLP_FW_VERSION, 0,
                                 sizeof(*cfm))) < 0)
                return err;

        *code = WLP_VERSION(cfm->major, cfm->minor, cfm->patch);
        if (build && build_len)
                owl_os_strncpy(build, cfm->build, build_len);
        if (hw_rev && hw_rev_len)
                owl_os_strncpy(hw_rev, cfm->hw_rev, hw_rev_len);

        owl_dbg("*code:%u build:%s len:%d hw_rev:%s len:%d", *code, build,
                build_len, hw_rev, hw_rev_len);
        return err;
}


void
_wlp_set_link_cb(void (*link_cb)(void *ctx, int link), void *ctx)
{
        owl_dbg("link_cb:%p ctx:%p", link_cb, ctx);
        
        priv->link = link_cb;
        priv->link_ctx = ctx;
}

int
_wlp_scan(struct wlp_ssid *ssid,
             void (*scan_cb)(void *ctx, struct wlp_network *net), void *ctx)
{
        struct wlp_scan_req *req = &priv->msg.u.req.u.scan;

        owl_os_memset(req, 0, sizeof(*req));
        priv->scan = scan_cb;
        priv->scan_ctx = ctx;
        
        if (ssid)
                owl_os_memcpy(&req->ssid, ssid, sizeof(req->ssid));

        return send_mgmt_req(&priv->msg, WLP_SCAN, sizeof(*req), 0);
}

int
_wlp_linkup(struct wlp_ssid *ssid, const char *key, enum wlp_auth_mode auth)
{
        struct wlp_link_up_req *req = &priv->msg.u.req.u.link_up;

        owl_dbg("ssid:%s key:%s auth:%d", ssid, key ? key : "null", auth);
        
        owl_os_memset(req, 0, sizeof(*req));

        if (ssid)
                owl_os_memcpy(&req->ssid, ssid, sizeof(req->ssid));
        
        if (key && auth != WLP_AUTH_WEP) {
                int len = owl_os_strlen(key);
                if (len < 8 || len > 63)
                        return WLP_ERR_PARAM;
        }
        if (key)
                owl_os_strncpy(req->key, key, sizeof(req->key) - 1);
        
        req->auth = (uint8_t) auth;
        
        return send_mgmt_req(&priv->msg, WLP_LINK_UP, sizeof(*req), 0);
}


int
_wlp_linkdown(void)
{
        owl_dbg("entry");
        
        return send_mgmt_req(&priv->msg, WLP_LINK_DOWN, 0, 0);
}


int
_wlp_set_ipaddr(const struct ip_addr *ip, const struct ip_addr *netmask,
                   const struct ip_addr *gw, const struct ip_addr *dns)
{
        owl_dbg("ip:%s netmask:%s gw:%s dns:%s",
                ip ? inet_ntoa(ip) : "NULL",
                netmask ? inet_ntoa(netmask) : "NULL",
                gw ? inet_ntoa(gw) : "NULL",
                dns ? inet_ntoa(dns) : "NULL");
        
        return net_config(ip, netmask, gw, dns, 0);
}


int
_wlp_get_ipaddr(struct ip_addr *ip, struct ip_addr *netmask,
                   struct ip_addr *gw, struct ip_addr *dns)
{
        owl_dbg("ip:%p netmask:%p gw:%p dns:%p", ip, netmask, gw, dns);

        return net_info(ip, netmask, gw, dns, NULL, NULL, NULL);
}

int
_wlp_set_dhcp(int enable)
{
        owl_dbg("enable:%d", enable);
        
        return net_config(NULL, NULL, NULL, NULL, enable);
}


int
_wlp_get_dhcp(int *enabled)
{
        owl_dbg("enabled:%p", enabled);
        return net_info(NULL, NULL, NULL, NULL, enabled, NULL, NULL);
}

int
_wlp_set_dhcpd(int enable)
{
        struct wlp_net_dhcpd_req *req = &priv->msg.u.req.u.net_dhcpd;

        owl_dbg("enable:%d", enable);
        req->enable = enable;
        return send_mgmt_req(&priv->msg, WLP_NET_DHCPD, sizeof(*req), 0);
}

void
_wlp_set_ipaddr_cb(void (*addr_cb)(void *ctx, const struct ip_addr *addr),
                       void *ctx)
{
        owl_dbg("addr_cb:%p ctx:%p", addr_cb, ctx);
        priv->addr = addr_cb;
        priv->addr_ctx = ctx;
}


int
_wlp_get_hwaddr(struct wlp_mac_addr *hwaddr)
{
        struct wlp_link_info_cfm *cfm = &priv->msg.u.cfm.u.link_info;
        int err;

        owl_dbg("hwaddr:%p", hwaddr);
        if (hwaddr == NULL)
                return WLP_ERR_PARAM;

        err = send_mgmt_req(&priv->msg, WLP_LINK_INFO, 0, sizeof(*cfm));
        if (err < 0 && err != WLP_ERR_CONN)
                return err;

        owl_os_memcpy(hwaddr, &cfm->hwaddr, sizeof(*hwaddr));
        owl_dbg("hwaddr:%02x%02x%02x%02x%02x%02x",
                hwaddr->octet[0], hwaddr->octet[1],
                hwaddr->octet[2], hwaddr->octet[3],
                hwaddr->octet[4], hwaddr->octet[5]);

        return 0;
}


int
_wlp_get_network(struct wlp_network *net)
{
        struct wlp_link_info_cfm *cfm = &priv->msg.u.cfm.u.link_info;
        int err;
        
        owl_dbg("net:%p", net);
        if (net == NULL)
                return WLP_ERR_PARAM;
        if ((err = send_mgmt_req(&priv->msg, WLP_LINK_INFO, 0, sizeof(*cfm)))
            < 0)
                return err;

        owl_os_memcpy(&net->bssid, &cfm->ap, sizeof(net->bssid));
        owl_os_memcpy(&net->ssid, &cfm->ssid, sizeof(net->ssid));
        net->rssi = ntohs(cfm->rssi);
        net->snr = ntohs(cfm->snr);
        owl_dbg("net->ssid:%s net->bssid:%02x%02x%02x%02x%02x%02x "
                "net->rssi:%d net->snr:%d",
                net->ssid.ssid,
                net->bssid.octet[0], net->bssid.octet[1], 
                net->bssid.octet[2], net->bssid.octet[3], 
                net->bssid.octet[4], net->bssid.octet[5],
                net->rssi, net->snr);

        net->channel = 0;
        net->auth = WLP_AUTH_AUTO;
        net->type = WLP_CONN_TYPE_INFRA;
        
        return 0;
}


int
_wlp_socket(int type, int proto)
{
        struct wlp_sock_open_req *req = &priv->msg.u.req.u.sock_open;
        struct wlp_sock_t *sock;
        int err;
        
        owl_dbg("type:%d proto:%d", type, proto);
        if ((sock = sock_alloc()) == NULL)
                return WLP_ERR_MEM;
        
        req->stype = type;
        req->protocol = proto;
        
        if ((err = send_mgmt_req(&priv->msg, WLP_SOCK_OPEN,
                                 sizeof(*req), 0)) < 0) {
                sock_free(sock);
                return err;
        }
                
        sock->id = err;
        sock->next = NULL;

        /* add socket to list */
        if (sock_head == NULL) {
                sock_head = sock;

        } else {
                struct wlp_sock_t *iter;
                for (iter = sock_head; iter->next; iter = iter->next);
                iter->next = sock;
        }

        owl_dbg("sock->id:%d", sock->id);
        return sock->id;
}


int
_wlp_set_conn_cb(int sockid,
                 void (*conn_cb)(void *ctx, int sockid, int connected),
                 void *ctx)
{
        struct wlp_sock_t *sock;

        owl_dbg("sockid:%d conn_cb:%p ctx:%p", sockid, conn_cb, ctx);
        if ((sock = sock_find(sockid)) == NULL)
                return WLP_ERR_PARAM;

        sock->conn = conn_cb;
        sock->ctx = ctx;
        return 0;
}


int
_wlp_set_recv_cb(int sockid,
                 void (*recv_cb)(void *ctx, int sockid, int len),
                 void *ctx)
{
        struct wlp_sock_t *sock;

        owl_dbg("sockid:%d recv_cb:%p ctx:%p", sockid, recv_cb, ctx);
        if ((sock = sock_find(sockid)) == NULL)
                return WLP_ERR_PARAM;

        sock->rx = recv_cb;
        sock->rx_ctx = ctx;
        return 0;
}

int
_wlp_bind(int sockid, const struct ip_addr *ip, uint16_t port)
{
        struct wlp_sock_bind_req *req = &priv->msg.u.req.u.sock_bind;

        owl_dbg("sockid:%d ip:%s port:%d",
                sockid, ip ? inet_ntoa(ip) : "NULL", port);
        
        req->sockid = htons(sockid);
        req->ip = *ip;
        req->port = htons(port);
        
        return send_mgmt_req(&priv->msg, WLP_SOCK_BIND, sizeof(*req), 0);
}


int
_wlp_listen(int sockid, int backlog,
            void (*listen_cb)(void *ctx, int sockid),
            void *ctx)
{
        struct wlp_sock_listen_req *req = &priv->msg.u.req.u.sock_listen;
        struct wlp_sock_t *sock;

        owl_dbg("sockid:%d backlog:%d listen_cb:%p ctx:%p",
                sockid, backlog, listen_cb, ctx);

        if ((sock = sock_find(sockid)) == NULL)
                return WLP_ERR_PARAM;

        sock->listen = listen_cb;
        sock->ctx = ctx;
        sock->flags = 0; /* clear any error flag */
        
        req->sockid = htons(sockid);
        req->backlog = backlog;
        
        return send_mgmt_req(&priv->msg, WLP_SOCK_LISTEN, sizeof(*req), 0);
}


int
_wlp_connect(int sockid, const struct ip_addr *ip, uint16_t port)
{
	struct wlp_sock_connect_req *req = &priv->msg.u.req.u.sock_connect;
	struct wlp_sock_t *sock;

        owl_dbg("sockid:%d ip:%s port:%d",
                sockid, ip ? inet_ntoa(ip) : "NULL", port);
                
        if ((sock = sock_find(sockid)) == NULL)
                return WLP_ERR_PARAM;

        sock->flags = 0; /* clear any error flag */
        
        req->sockid = htons(sockid);
        req->ip = *ip;
        req->port = htons(port);
        
        return send_mgmt_req(&priv->msg, WLP_SOCK_CONNECT, sizeof(*req), 0);
}


int
_wlp_accept(int sockid)
{
        struct wlp_sock_accept_req *req = &priv->msg.u.req.u.sock_accept;
        struct wlp_sock_t *sock;
        struct wlp_sock_t *iter;
        int err;

        owl_dbg("sockid:%d", sockid);

        if ((sock = sock_alloc()) == NULL)
                return WLP_ERR_MEM;

        req->sockid = htons(sockid);
        
        if ((err = send_mgmt_req(&priv->msg, WLP_SOCK_ACCEPT,
                                 sizeof(*req), 0)) < 0) {
                sock_free(sock);
                return err;
        }
        
        sock->id = err;
        sock->next = NULL;
        sock->flags |= WLP_SOCK_FLAG_CONNECTED;
        
        /* sock_head always != NULL, since we have at least our listening
         * sock
         */
        for (iter = sock_head; iter->next; iter = iter->next);
        iter->next = sock;
        
        return sock->id;
}


int
_wlp_close(int sockid)
{
        struct wlp_sock_close_req *req = &priv->msg.u.req.u.sock_close;
        struct wlp_sock_t *iter;
        struct wlp_sock_t *prev = NULL;

        owl_dbg("sockid:%d", sockid);

        req->sockid = htons(sockid);

        /* make sure that we delete the socket from our local list even though
         * the request fails. This might happen if the device was restarted
         * with wlp_reset() while we still had sockets open
         */
        (void) send_mgmt_req(&priv->msg, WLP_SOCK_CLOSE, sizeof(*req), 0);
        
        for (iter = sock_head; iter; prev = iter, iter = iter->next) {
                if (iter->id != sockid)
                        continue;

                /*
                 * Already a pending close for this socket? Maybe we are trying
                 * to close another socket with the same id. Note that there
                 * is a possibility that we have two or more sockets in our
                 * sock_head list that share the same id's. This will happen
                 * if wlp_close() is closed from any callbacks invoked from
                 * wlp_poll() _and_ wlp_connect() or wlp_accept() is called
                 * before wlp_poll() has returned.
                 *
                 */
                if (iter->flags & WLP_SOCK_FLAG_DELETED)
                        continue;
                
                /* invoked from a wlp callback? then just mark the socket
                 * as deleted, and it will be removed when polling is complete,
                 * from wlp_poll()
                 */
                if (priv->flags & WLP_FLAG_POLL) {
                        iter->flags |= WLP_SOCK_FLAG_DELETED;
                        return 0;
                }

                if (prev == NULL)
                        sock_head = sock_head->next;
                else
                        prev->next = iter->next;
                
                sock_free(iter);
                return 0;
        }

        return 0;
}


int
_wlp_recv(int sockid, char *buf, int16_t len)
{
        struct wlp_sock_recv_req *req = &priv->msg.u.req.u.sock_recv;
        int err;

        owl_dbg("sockid:%d buf:%p len:%d", sockid, buf, len);
        
        req->sockid = htons(sockid);
        req->datalen = htons(len);
        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 1);

        bus_claim();

        if ((err = send_mgmt_req_ext(&priv->msg, WLP_SOCK_RECV,
                                     sizeof(*req), 0, 0)) < 0) {
                goto out;
        }

        /* should never happen */
        if (err > len) {
                owl_err("too many bytes:%d len:%d", err, len);
                err = WLP_ERR_PROTO;
                goto out;
        }
                        
        /* read result number of bytes into buf */
        err = recv_data_ind(buf, err);
out:
        bus_release();
        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 0);

        return err;
}


int
_wlp_send(int sockid, const char *buf, int16_t len)
{
        struct wlp_sock_send_req *req = &priv->msg.u.req.u.sock_send;
        int err;

        owl_dbg("sockid:%d buf:%p len:%d", sockid, buf, len);
        
        req->sockid = htons(sockid);
        req->datalen = htons(len);

        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 1);
        bus_claim();

        if ((err = send_mgmt_req_ext(&priv->msg, WLP_SOCK_SEND,
                                     sizeof(*req), 0, 0)) < 0)
                goto out;

        /* should never happen */
        if (err > len) {
                err = WLP_ERR_PROTO;
                goto out;
        }

        /* write result number of bytes from buf */
        err = send_data_req(buf, err);
out:
        bus_release();
        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 0);

        return err;
}


int
_wlp_sendto(int sockid, const char *buf, int16_t len,
               const struct ip_addr *ip, uint16_t port)
{
        struct wlp_sock_sendto_req *req = &priv->msg.u.req.u.sock_sendto;
        int err;

        owl_dbg("sockid:%d buf:%p len:%d ip:%s port:%d",
                sockid, buf, len, ip ? inet_ntoa(ip) : "NULL", port);
        
        req->sockid = htons(sockid);
        req->datalen = htons(len);
        req->addr = *ip;
        req->port = htons(port);
        
        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 1);
        bus_claim();

        if ((err = send_mgmt_req_ext(&priv->msg, WLP_SOCK_SENDTO,
                                     sizeof(*req), 0, 0)) < 0)
                goto out;

        /* should never happen */
        if (err > len) {
                err = WLP_ERR_PROTO;
                goto out;
        }

        /* write result number of bytes from buf */
        err = send_data_req(buf, err);
out:
        bus_release();
        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 0);

        return err;
}


int
_wlp_get_peeraddr(int sockid, struct ip_addr *peer)
{
        struct wlp_sock_info_req *req = &priv->msg.u.req.u.sock_info;
        struct wlp_sock_info_cfm *cfm = &priv->msg.u.cfm.u.sock_info;
        int err;

        owl_dbg("sockid:%d", sockid); 

        req->sockid = htons(sockid);
        
        if ((err = send_mgmt_req(&priv->msg, WLP_SOCK_INFO,
                            sizeof(*req), sizeof(*cfm))) < 0)
                return err;
        
        *peer = cfm->peer;
        owl_dbg("*peer:%s", inet_ntoa(peer));
        
        return err;              
}


int
_wlp_get_hostbyname(const char *host,
		    void (*lookup_cb)(void *ctx, const struct ip_addr *ip),
		    void *ctx)
{
        struct wlp_net_dnslookup_req *req = &priv->msg.u.req.u.net_dnslookup;
        int err;
        
        owl_dbg("host:%s lookup_cb:%p ctx:%p", host, lookup_cb, ctx); 
        
        if (host == NULL)
                return WLP_ERR_PARAM;
        
        if (owl_os_strlen(host) > WLP_HOSTNAME_MAX_LENGTH)
                return WLP_ERR_PARAM;

        if (priv->lookup)
                return WLP_ERR_BUSY;
        
        owl_os_strcpy(req->host, host);
        
        if ((err = send_mgmt_req(&priv->msg, WLP_NET_DNSLOOKUP,
                                 sizeof(*req), 0)) < 0)
                return err;

        priv->lookup = lookup_cb;
        priv->lookup_ctx = ctx;
        return err;              
}



int
_wlp_poll(void)
{
        int res;
        struct ip_addr ip;
        struct ip_addr dns;
        int dns_result;
        struct wlp_sock_t *prev = NULL;
        struct wlp_sock_t *iter;
        int event;
        
        if (priv == NULL)
                return 0;
        
        dbg_inhibit(); /* disable debugging to not overflow the debug output */
        event = send_mgmt_req(&priv->msg, WLP_POLL, 0, 0);
        dbg_allow();

        /* no event? then no need to check anything */
        if (!event)
                return 0;

        /* check wifi link status, invoke link_cb accordingly */
        res = send_mgmt_req(&priv->msg, WLP_LINK_STATUS, 0, 0);
        if (res == 0 && priv->link && !(priv->flags & WLP_FLAG_LINKUP)) {
                priv->flags |= WLP_FLAG_LINKUP;
                owl_info("flags:%x", priv->flags);
                priv->link(priv->link_ctx, 1);
                
        } else if (res < 0 && priv->link && (priv->flags & WLP_FLAG_LINKUP)) {
                priv->flags &= ~(WLP_FLAG_LINKUP);
                owl_info("flags:%x", priv->flags);
                priv->link(priv->link_ctx, 0);
        }

        if (priv->scan) {
                struct wlp_scan_status_cfm *cfm =
                        &priv->msg.u.cfm.u.scan_status;

                for (;;) {
                        res = send_mgmt_req(
                                &priv->msg, WLP_SCAN_STATUS, 0, sizeof(*cfm));
                        
                        if (res == 0) {
                                priv->scan(priv->scan_ctx, NULL);
                                priv->scan = NULL;
                        } else if (res > 0) {
                                cfm->net.rssi = ntohs(cfm->net.rssi);
                                cfm->net.snr = ntohs(cfm->net.snr);
                                priv->scan(priv->scan_ctx, &cfm->net);
                                continue;
                        }

                        break;
                } 
        }
        
        /* check network status, invoke addr_cb */
        res = net_info(&ip, NULL, NULL, NULL, NULL, &dns, &dns_result);
        if (res == 0 && priv->addr && !(priv->flags & WLP_FLAG_GOT_ADDR)) {
                priv->flags |= WLP_FLAG_GOT_ADDR;
                owl_info("flags:%x", priv->flags);
                priv->addr(priv->addr_ctx, &ip);

        } else if (res < 0 && priv->addr && (priv->flags & WLP_FLAG_GOT_ADDR)) {
                priv->flags &= ~(WLP_FLAG_GOT_ADDR);
                owl_info("flags:%x", priv->flags);
                priv->addr(priv->addr_ctx, NULL);
        }
                
        /* invoke any dns lookup cb's */
        if (res == 0 && priv->lookup) {
                if (dns_result == WLP_DNS_OK) {
                        owl_info("lookup complete, dns:%s", inet_ntoa(&dns));
                        priv->lookup(priv->lookup_ctx, &dns);
                        priv->lookup = NULL;
                } else if (dns_result == WLP_DNS_ERROR) {
                        owl_info("lookup error");
                        priv->lookup(priv->lookup_ctx, NULL);
                        priv->lookup = NULL;
                }
        }

        /* invoke linkctx cb */
        if (priv->linkctx) {
                struct wlp_net_info_cfm *cfm = &priv->msg.u.cfm.u.net_info;
                if (cfm->linkctx_len) {
                        priv->linkctx(priv->linkctx_ctx,
                                      cfm->linkctx,
                                      cfm->linkctx_len); 
  
                        priv->linkctx = NULL;
                        priv->linkctx_ctx = NULL;
                }
        }

        /* Check device status for fw upgrade status */
        if ((priv->flags & WLP_FLAG_FW_UPGRADE) && priv->fw_up_status) {
                uint8_t dev_flags;
                dev_flags = get_dev_status();
                if (dev_flags & WLP_DEV_STATUS_FW_SUCCESS) {
                        priv->fw_up_status(priv->fw_up_status_ctx, 1);
                }
                else if (dev_flags & WLP_DEV_STATUS_FW_FAIL) {
                        priv->fw_up_status(priv->fw_up_status_ctx, 0);
                }
        }

        /* to avoid closing sockets while iterating through them */
        priv->flags |= WLP_FLAG_POLL;

        /* check status for each sock, invoke conn cb or rx cb accordingly */
        for (iter = sock_head; iter; iter = iter->next) {
                int avail;
                int8_t state;

                avail = sock_status(iter->id, &state);
                
                if (avail < 0)
                        /* fail to get status, consider this socket lost.
                         * This might happen after wlp_reset() if some sockets
                         *  were open prior to the call
                         */
                        state = WLP_SOCK_STATE_ERROR;
                
                if (state == WLP_SOCK_STATE_CONNECTED &&
                    !(iter->flags & WLP_SOCK_FLAG_CONNECTED) && iter->conn) {
                        /* got connection */
                        iter->flags |= WLP_SOCK_FLAG_CONNECTED;
                        iter->flags &= ~WLP_SOCK_FLAG_ERROR;
                        owl_info("flags:%x", priv->flags);
                        iter->conn(iter->ctx, iter->id, 1);
                        
                } else if (state == WLP_SOCK_STATE_ERROR &&
                           !(iter->flags & WLP_SOCK_FLAG_ERROR) &&
                           iter->conn) {
                        /* connection error */
                        iter->flags &= ~WLP_SOCK_FLAG_CONNECTED;
                        iter->flags |= WLP_SOCK_FLAG_ERROR;
                        owl_info("flags:%x", priv->flags);
                        iter->conn(iter->ctx, iter->id, 0);
                        
                } else if (state == WLP_SOCK_STATE_ACCEPT && iter->listen) {
                        /* got incoming connection */
                        owl_info("accept id:%d", iter->id);
                        iter->listen(iter->ctx, iter->id);
                        
                } else if (avail > 0 && iter->rx) {
                        /* data pending, we check this before checking the
                         * state to make sure the client gets all the data
                         * before the rx_cb() signals a disconnection through
                         * len 0 (next case)
                         */
                        owl_info("rx id:%d avail:%d", iter->id, avail);
                        iter->rx(iter->rx_ctx, iter->id, avail);
                        
                } else if (state == WLP_SOCK_STATE_DISCONNECTED &&
                           (iter->flags & WLP_SOCK_FLAG_CONNECTED) &&
                           iter->rx) {
                        /* closed conn */
                        iter->flags &= ~WLP_SOCK_FLAG_CONNECTED;
                        owl_info("rx id:%d avail:%d", iter->id, 0);
                        iter->rx(iter->rx_ctx, iter->id, 0);
                }
        }

        priv->flags &= ~WLP_FLAG_POLL;

        /* if wlp_close() was called from a callback, we need to finalize the
         * closing of those sockets here.
         */
        iter = sock_head;
        while (iter) {
                struct wlp_sock_t *sock = iter;
                if (!(iter->flags & WLP_SOCK_FLAG_DELETED))
                        goto loop;

                if (prev == NULL)
                        sock_head = sock_head->next;
                else
                        prev->next = iter->next;

                iter = iter->next;
                sock_free(sock);
                continue;
        loop:
                prev = iter;
                iter = iter->next;
        }

        return 0;
}

int
_wlp_set_wps(int enable)
{
        struct wlp_set_wps_req *req = &priv->msg.u.req.u.wps;

        owl_dbg("enable:%d", enable);
        
        req->enable = enable;
        
        return send_mgmt_req(&priv->msg, WLP_SET_WPS, sizeof(*req), 0);
}

int
_wlp_get_link_context(void (*linkctx_cb)(void *ctx, void *linkctx, int len),
		      void *ctx)
{
        if (priv->linkctx)
                return WLP_ERR_PARAM;
        
        priv->linkctx = linkctx_cb;
        priv->linkctx_ctx = ctx;

        return send_mgmt_req(&priv->msg, 
                             WLP_GET_LINKCTX,
                             0, 0);
}

int
_wlp_set_link_context(void *linkctx)
{
        struct wlp_set_linkctx_req *req = &priv->msg.u.req.u.set_linkctx;

        owl_os_memcpy(req, linkctx, WLP_LINKCTX_MAX_LEN);
                return send_mgmt_req(&priv->msg,
                                     WLP_SET_LINKCTX,
                                     sizeof(*req),
                                     0);
}

int
_wlp_set_mode(uint8_t mode)
{
        struct wlp_set_mode_req *req = &priv->msg.u.req.u.set_mode;

        owl_dbg("mode:%d", mode);
        
        req->mode = mode;
        
        return send_mgmt_req(&priv->msg, WLP_SET_MODE, sizeof(*req), 0);
}


int
_wlp_get_mode(uint8_t *mode)
{
        struct wlp_get_mode_cfm *cfm = &priv->msg.u.cfm.u.get_mode;
        int err;

        owl_dbg("mode:%p", mode);
        
        if ((err = send_mgmt_req(&priv->msg, WLP_GET_MODE,
                                 0, sizeof(*cfm))) < 0)
                return err;

        *mode = cfm->mode;
        return err;
}


int
_wlp_set_channel(uint8_t channel,
                 void (*channel_cb)(void *ctx, int status), void *ctx)
{
        struct wlp_set_channel_req *req = &priv->msg.u.req.u.set_channel;

        owl_dbg("channel:%d", channel);
        req->channel = channel;
        return send_mgmt_req(&priv->msg, WLP_SET_CHANNEL, sizeof(*req), 0);
}

int
_wlp_conf_ap(int hidden, int beacon_intvl, int inact_tmo,
                const char *country)
{
        return WLP_ERR_NOTSUPP;
}

int
_wlp_set_ps(int enable)
{
        struct wlp_set_ps_enable_req *req = &priv->msg.u.req.u.ps_enable;

        req->enable = !! enable;

        return send_mgmt_req(&priv->msg, WLP_SET_PS_ENABLE, sizeof(*req), 0);
}

int
_wlp_conf_ps(uint8_t ps_poll,
	     uint32_t traffic_timeout,
	     uint32_t ps_delay,
	     uint8_t rx_all_dtim,
	     uint16_t listen_interval)
{
        struct wlp_set_ps_conf_req *req = &priv->msg.u.req.u.ps_config;

        owl_dbg("ps_delay:%d traffic_timout:%d "
                "listen_interval:%d ps_poll:%d rx_all_dtim:%d)",
                ps_delay,
                traffic_timeout,
                listen_interval,
                ps_poll,
                rx_all_dtim);
        req->ps_delay = htonl(ps_delay);
        req->traffic_timeout = htonl(traffic_timeout);
        req->listen_interval = htons(listen_interval);
        req->ps_poll = ps_poll;
        req->rx_all_dtim = rx_all_dtim;
        return send_mgmt_req(&priv->msg, WLP_SET_PS_CONF, sizeof(*req), 0);
}

int
_wlp_fw_upgrade(uint8_t method, 
		struct ip_addr *addr,
		uint16_t port,
		const char *path) 
{
        struct wlp_fw_upgrade_req *req = &priv->msg.u.req.u.fw_upgrade;
        int err;

        owl_os_memset(req, 0, sizeof(*req));
        if (addr)
                req->addr = *addr;
        req->port = htons(port);
        req->method = method;
        if (path) {
                if (owl_os_strlen(path) < WLP_PATH_MAX_LENGTH)
                        owl_os_memcpy(req->path, path, owl_os_strlen(path));
                else
                        return WLP_ERR_PARAM;
        }         
        err = send_mgmt_req(&priv->msg, WLP_FW_UPGRADE, sizeof(*req), 0);
        if (0 == err)
                priv->flags |= WLP_FLAG_FW_UPGRADE;
        return err;
}

int
_wlp_set_fw_upgrade_cb(void (*status_cb)(void *ctx, uint8_t success),
		       void *ctx) {
	
        priv->fw_up_status = status_cb;
        priv->fw_up_status_ctx = ctx;

        return 0;
}

int
_wlp_set_led(int enable)
{
        struct wlp_set_led_req *req = &priv->msg.u.req.u.led_enable;
	
        req->enable = !! enable;

        return send_mgmt_req(&priv->msg, WLP_SET_LED, sizeof(*req), 0);

}

int
_wlp_set_heartbeat(uint32_t ms_period)
{
        struct wlp_set_heartbeat_req *req = &priv->msg.u.req.u.heartbeat;

        req->period = htonl(ms_period);

        return send_mgmt_req(&priv->msg, WLP_SET_HEARTBEAT, sizeof(*req), 0);

}

int
_wlp_set_poll_period(uint32_t ms_period)
{
        struct wlp_set_poll_period_req *req = &priv->msg.u.req.u.poll_period;

        req->period = htonl(ms_period);

        return send_mgmt_req(&priv->msg, WLP_SET_POLL_PERIOD, sizeof(*req), 0);
}

static uint8_t
get_dev_status(void)
{
        struct wlp_dev_status_cfm *cfm = &priv->msg.u.cfm.u.dev_status;
        uint8_t flags;
        int err;

        if ((err = send_mgmt_req(&priv->msg, WLP_DEV_STATUS,
                                 0, sizeof(*cfm))) < 0)
                return 0;

        flags = cfm->flags;
        return flags;
}

static int
net_config(const struct ip_addr *ip, const struct ip_addr *netmask,
	   const struct ip_addr *gw, const struct ip_addr *dns,
	   int dhcp)
{
        struct wlp_net_config_req *req = &priv->msg.u.req.u.net_config;
     
        req->ip  = ip ? *ip : ip_addr_any;
        req->netmask = netmask ? *netmask : ip_addr_any;
        req->gw = gw ? *gw : ip_addr_any;
        req->dns_server = dns ? *dns : ip_addr_any;
        req->dhcp = dhcp;
        
        return send_mgmt_req(&priv->msg, WLP_NET_CONFIG, sizeof(*req), 0);
}


static int
net_info(struct ip_addr *ip, struct ip_addr *netmask,
	 struct ip_addr *gw, struct ip_addr *dns, int *dhcp,
	 struct ip_addr *dns_lookup, int *dns_result)
{
        struct wlp_net_info_cfm *cfm = &priv->msg.u.cfm.u.net_info;
        int err;

        owl_dbg("entry");

        err = send_mgmt_req(&priv->msg, WLP_NET_INFO, 0, sizeof(*cfm));
        
        /* fill in parameters also when not connected */
        
        if (ip)
                *ip = cfm->ip;

        if (netmask)
                *netmask = cfm->netmask;

        if (gw)
                *gw = cfm->gw;

        if (dns)
                *dns = cfm->dns_server;

        if (dhcp)
                *dhcp = cfm->dhcp;

        if (dns_result)
                *dns_result = cfm->dns_result;

        if (dns_lookup)
                *dns_lookup = cfm->dns_lookup;
        
        owl_dbg("ip:%s netmask:%s gw:%s dns:%s dhcp:%d ",
                inet_ntoa(&cfm->ip), inet_ntoa(&cfm->netmask),
                inet_ntoa(&cfm->gw), inet_ntoa(&cfm->dns_server),
                cfm->dhcp);
        owl_dbg("dns_lookup:%s dns_result:%d",
                inet_ntoa(&cfm->dns_lookup), cfm->dns_result);
        
        return err;
}


static int
sock_status(int16_t sockid, int8_t *state)
{
        struct wlp_sock_status_req *req = &priv->msg.u.req.u.sock_status;
        struct wlp_sock_status_cfm *cfm = &priv->msg.u.cfm.u.sock_status;
        int err;

        req->sockid = htons(sockid);
        
        if ((err = send_mgmt_req(&priv->msg, WLP_SOCK_STATUS,
                            sizeof(*req), sizeof(*cfm))) < 0)
                return err;

        *state = cfm->state;
        
        owl_dbg("state:%d", cfm->state);
        return err;
}
     
static int
send_mgmt_req(struct wlp_msg *msg, int id, int reqlen, int cfmlen)
{
        int ret;
        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 1);

        ret =  send_mgmt_req_ext(msg, id, reqlen, cfmlen, 1);

        if (priv->wakeup) 
                priv->wakeup(priv->wakeup_ctx, 0);

        return ret;
}


static int
send_mgmt_req_ext(struct wlp_msg *msg, int id, int reqlen,
		  int cfmlen, int claimbus)
{
        dbg_allowed(owl_dbg("msg:%p id:%d reqlen:%d cfmlen:%d claimbus:%d",
                            msg, id, reqlen, cfmlen, claimbus));
        
        reqlen += sizeof(struct wlp_hdr);
        cfmlen += sizeof(struct wlp_hdr);

        msg->hdr.len = htons(reqlen);
        msg->hdr.type = WLP_TYPE_MGMT_REQ;
        msg->hdr.id = id;
        msg->hdr.res = 0;
        msg->hdr.pad = 0;

        dbg_allowed(owl_hexdump("req", (char *) msg, reqlen));
        
        if (claimbus) {
                bus_claim();
                bus_write(msg, reqlen);
                bus_wait();
                bus_read(msg, cfmlen);
                bus_release();
        } else {
                bus_write(msg, reqlen);
                bus_wait();
                bus_read(msg, cfmlen);
        }

        dbg_allowed(owl_hexdump("cfm", (char *) msg, cfmlen));

        msg->hdr.len = ntohs(msg->hdr.len);
        msg->hdr.res = ntohs(msg->hdr.res);
        
        if (msg->hdr.len != cfmlen) {
                owl_err("invalid len hdr.len:%d cfmlen:%d",
                        msg->hdr.len, cfmlen);
                return WLP_ERR_PROTO;
                
        } else if (msg->hdr.type != WLP_TYPE_MGMT_CFM) {
                owl_err("invalid type:%d expected:%d",
                        msg->hdr.type, WLP_TYPE_MGMT_CFM);
                return WLP_ERR_PROTO;
                
        } else if (msg->hdr.id != id) {
                owl_err("invalid id:%d expected:%d", msg->hdr.id, id);
                return WLP_ERR_PROTO;
        }

        dbg_allowed(owl_dbg("res:%d", msg->hdr.res));
        return msg->hdr.res;
}

static int
send_data_req(const char *data, int len)
{
        int totlen;
        int err;
        struct wlp_hdr *hdr = &priv->msg.hdr;

        owl_dbg("data:%p len:%d", data, len);

        /* If the length is zero (because the server could
           not accept any more data right now) then we skip
           the data request. */
        if (0 == len) 
                return 0;

        /* fill in the header */
        hdr->pad = wlp_pad(len);
        totlen = len + sizeof(*hdr) + hdr->pad;
        hdr->len = htons(totlen);
        hdr->type = WLP_TYPE_DATA_REQ;
        hdr->id = WLP_SOCK_DATA;
        hdr->res = 0;

        owl_dbg("hdr->len:%d hdr->pad:%d hdr->type:%d hdr->id:%d",
                ntohs(hdr->len), hdr->pad, hdr->type, hdr->id);
        owl_hexdump("tx payload", data, len);

        /* send the header */
        bus_write(hdr, sizeof(*hdr));

        /* Send the data. NOTE: we will read outside the buffer to support
           padding without passing odd-sized data to the bus driver */
        len += hdr->pad;
        err = bus_write(data, len); 
        if (err != len) {
                owl_err("write failed err:%d len:%d", err, len);
                return WLP_ERR_PROTO;
        }

        /* wait for device */
        bus_wait();

        /* receive data cfm */
        bus_read(hdr, sizeof(*hdr));

        hdr->len = ntohs(hdr->len);
        hdr->res = ntohs(hdr->res);
        
        if (hdr->len != sizeof(*hdr)) {
                owl_err("invalid len hdr->len:%d sizeof(*hdr):%d",
                        hdr->len, sizeof(*hdr));
                return WLP_ERR_PROTO;
                
        } else if (hdr->type != WLP_TYPE_DATA_CFM) {
                owl_err("invalid type:%d expected:%d",
                        hdr->type, WLP_TYPE_DATA_CFM);
                return WLP_ERR_PROTO;
                
        } else if (hdr->id != WLP_SOCK_DATA) {
                owl_err("invalid id:%d expected:%d", hdr->id, WLP_SOCK_DATA);
                return WLP_ERR_PROTO;
        }

        owl_dbg("res:%d", hdr->res);
        return hdr->res;
}

static int
recv_data_ind(char *data, int len)
{
        int err;
        char pad[3];
        struct wlp_hdr *hdr = &priv->msg.hdr;
        
        owl_dbg("data:%p len:%d", data, len);

        if (0 == len)
                return 0;
        /* wait for device */
        bus_wait();
        
        /* recv the header */
        err = bus_read(hdr, sizeof(*hdr));
        hdr->len = ntohs(hdr->len);
        
        if (err != sizeof(*hdr)) {
                owl_err("invalid read size:%d expected:%d", err, sizeof(*hdr));
                len = WLP_ERR_PROTO;
                goto out;
                
        } else if (hdr->type != WLP_TYPE_DATA_IND) {
                owl_err("invalid type:%d expected:%d",
                        hdr->type, WLP_TYPE_DATA_IND);
                len = WLP_ERR_PROTO;
                goto out;

        } else if (hdr->id != WLP_SOCK_DATA) {
                owl_err("invalid id:%d expected:%d",
                        hdr->id, WLP_SOCK_DATA);
                len = WLP_ERR_PROTO;
                goto out;
                
        } else if (hdr->len - hdr->pad - sizeof(*hdr) != len) {
                owl_err("invalid len/pad hdr->len:%d hdr->pad:%d len:%d",
                        hdr->len, hdr->pad, len);
                len = WLP_ERR_PROTO;
                goto out;
        }

        owl_dbg("hdr->len:%d hdr->pad:%d hdr->type:%d hdr->id:%d",
                hdr->len, hdr->pad, hdr->type, hdr->id);

        /* read the payload */
        err = bus_read(data, len);
        if (err != len) {
                owl_err("read failed err:%d len:%d", err, len);
                len = WLP_ERR_PROTO;
                goto out;
        }

        /* read any padding */
        if (hdr->pad)
                bus_read(pad, hdr->pad);
out:
        owl_hexdump("rx payload", data, len);
        
        return len;
}

static struct wlp_sock_t *
sock_find(int id)
{
        struct wlp_sock_t *iter;
        for (iter = sock_head; iter; iter = iter->next) {
                if (iter->flags & WLP_SOCK_FLAG_DELETED)
                        continue;
                if (iter->id == id)
                        return iter;
        }

        return NULL;
}


static struct wlp_sock_t *
sock_alloc(void)
{
        struct wlp_sock_t *sock;
        
#if WLP_HEAP
        sock = owl_os_zalloc(sizeof(struct wlp_sock_t));
        owl_dbg("sock:%p", sock);
        return sock;
        
#else 
        int i;
        for (i = 0; i < sizeof(priv->sock) / sizeof(priv->sock[0]); i++) {
                struct wlp_sock_t *sock = &priv->sock[i];
                if (!sock->id) {
                        owl_os_memset(sock, 0, sizeof(struct wlp_sock_t));
                        owl_dbg("sock:%p", sock);
                        return sock;
                }
        }

        owl_err("sock:NULL");
        return NULL;
#endif /* WLP_HEAP */
}        


static void
sock_free(struct wlp_sock_t *sock)
{
        owl_dbg("sock:%p sock->id:%d", sock, sock ? sock->id : 0);
        
#if WLP_HEAP
        owl_os_free(sock);
#else
        if (sock)
                sock->id = 0;
#endif /* WLP_HEAP */
}

static int
bus_read(void *data, int len)
{
        return priv->read(priv->rw_ctx, data, len);
}

static int
bus_write(const void *data, int len)
{
       return priv->write(priv->rw_ctx, data, len);
}

static void
bus_claim(void)
{
        if (priv->select)
                priv->select(priv->rw_ctx, 1);
}

static void
bus_release(void)
{
        if (priv->select)
                priv->select(priv->rw_ctx, 0);
}

static void
bus_wait(void)
{
        if (!(priv->flags & WLP_FLAG_SPI))
                return;
        
        while (!priv->irq) {
                if (priv->yield)
                        priv->yield(priv->yield_ctx);
        }

        priv->irq = 0;
}
